import math
from collections import defaultdict, deque
from typing import Any, Iterator, TypeAlias

import pandas as pd

from qubx import logger
from qubx.core.basics import DataType, Instrument, Timestamped
from qubx.core.exceptions import SimulationError
from qubx.data.readers import (
    AsDict,
    AsQuotes,
    AsTrades,
    DataReader,
    DataTransformer,
    RestoredBarsFromOHLC,
    RestoreQuotesFromOHLC,
    RestoreTradesFromOHLC,
)

SlicerOutData: TypeAlias = tuple[str, int, Timestamped] | tuple


class IteratedDataStreamsSlicer(Iterator[SlicerOutData]):
    """
    This class manages seamless iteration over multiple time-series data streams,
    ensuring that events are processed in the correct chronological order regardless of their source.
    It supports adding / removing new data streams to the slicer on the fly (during the itration).
    """

    _iterators: dict[str, Iterator[list[Timestamped]]]
    _buffers: dict[str, list[Timestamped]]
    _keys: deque[str]
    _iterating: bool

    def __init__(self):
        self._buffers = defaultdict(list)
        self._iterators = {}
        self._keys = deque()
        self._iterating = False

    def put(self, data: dict[str, Iterator[list[Timestamped]]]):
        _rebuild = False
        for k, vi in data.items():
            if k not in self._keys:
                self._iterators[k] = vi
                self._buffers[k] = self._load_next_chunk_to_buffer(k)  # do initial chunk fetching
                self._keys.append(k)
                _rebuild = True

        # - rebuild strategy
        if _rebuild and self._iterating:
            self._build_initial_iteration_seq()

    def __add__(self, data: dict[str, Iterator]) -> "IteratedDataStreamsSlicer":
        self.put(data)
        return self

    def remove(self, keys: list[str] | str):
        """
        Remove data iterator and associated keys from the queue.
        If the key is not found, it does nothing.
        """
        _keys = keys if isinstance(keys, list) else [keys]
        _rebuild = False
        for i in _keys:
            if i in self._buffers:
                self._buffers.pop(i)
                self._iterators.pop(i)
                self._keys.remove(i)
                _rebuild = True

        # - rebuild strategy
        if _rebuild and self._iterating:
            self._build_initial_iteration_seq()

    def __iter__(self) -> Iterator:
        self._build_initial_iteration_seq()
        self._iterating = True
        return self

    def _build_initial_iteration_seq(self):
        _init_seq = {k: self._buffers[k][-1].time for k in self._keys}
        _init_seq = dict(sorted(_init_seq.items(), key=lambda item: item[1]))
        self._keys = deque(_init_seq.keys())

    def _load_next_chunk_to_buffer(self, index: str) -> list[Timestamped]:
        return list(reversed(next(self._iterators[index])))

    def _remove_iterator(self, key: str):
        self._buffers.pop(key)
        self._iterators.pop(key)
        self._keys.remove(key)

    def _pop_top(self, k: str) -> Timestamped:
        """
        Removes and returns the most recent timestamped data element from the buffer associated with the given key.
        If the buffer is empty after popping, it attempts to load the next chunk of data into the buffer.
        If no more data is available, the iterator associated with the key is removed.

        Parameters:
            k (str): The key identifying the data stream buffer to pop from.

        Returns:
            Timestamped: The most recent timestamped data element from the buffer.
        """
        v = (data := self._buffers[k]).pop()
        if not data:
            try:
                data.extend(self._load_next_chunk_to_buffer(k))  # - get next chunk of data
            except StopIteration:
                self._remove_iterator(k)  # - remove iterable data
        return v

    def fetch_before_time(self, key: str, time_ns: int) -> list[Timestamped]:
        """
        Fetches and returns all timestamped data elements from the buffer associated with the given key
        that have a timestamp earlier than the specified time.

        Parameters:
            - key (str): The key identifying the data stream buffer to fetch from.
            - time_ns (int): The timestamp in nanoseconds. All returned elements will have a timestamp less than this value.

        Returns:
            - list[Timestamped]: A list of timestamped data elements that occur before the specified time.
        """
        values = []
        data = self._buffers[key]
        if not data:
            try:
                data.extend(self._load_next_chunk_to_buffer(key))  # - get next chunk of data
            except StopIteration:
                self._remove_iterator(key)

        # pull most past elements
        v = data[-1]
        while v.time < time_ns:
            values.append(data.pop())
            if not data:
                try:
                    data.extend(self._load_next_chunk_to_buffer(key))  # - get next chunk of data
                except StopIteration:
                    self._remove_iterator(key)
                    break
            v = data[-1]

        return values

    def __next__(self) -> SlicerOutData:
        """
        Advances the iterator to the next available timestamped data element across all data streams.

        Returns:
            - SlicerOutData: A tuple containing the key of the data stream, the timestamp of the data element, and the data element itself.

        Raises:
            - StopIteration: If there are no more data elements to iterate over.
        """
        if not self._keys:
            self._iterating = False
            raise StopIteration

        _min_t = math.inf
        _min_k = self._keys[0]
        for i in self._keys:
            _x = self._buffers[i][-1]
            if _x.time < _min_t:
                _min_t = _x.time
                _min_k = i

        _v = self._pop_top(_min_k)
        return (_min_k, _v.time, _v)


class DataFetcher:
    _fetcher_id: str
    _reader: DataReader
    _requested_data_type: str
    _producing_data_type: str
    _params: dict[str, object]
    _specs: list[str]

    _transformer: DataTransformer
    _timeframe: str | None = None
    _warmup_period: pd.Timedelta | None = None
    _chunksize: int = 5000

    def __init__(
        self,
        fetcher_id: str,
        reader: DataReader,
        subtype: str,
        params: dict[str, Any],
        warmup_period: pd.Timedelta | None = None,
        chunksize: int = 5000,
        open_close_time_indent_secs=1.0,  # open/close shift may depends on simulation
    ) -> None:
        self._fetcher_id = fetcher_id
        self._params = params
        self._reader = reader

        match subtype:
            case DataType.OHLC_QUOTES:
                # - requested restore quotes from OHLC
                self._transformer = RestoreQuotesFromOHLC(open_close_time_shift_secs=open_close_time_indent_secs)
                self._requested_data_type = "ohlc"
                self._producing_data_type = "quote"
                if "timeframe" in params:
                    self._timeframe = params.get("timeframe", "1Min")

            case DataType.OHLC_TRADES:
                # - requested restore trades from OHLC
                self._transformer = RestoreTradesFromOHLC(open_close_time_shift_secs=open_close_time_indent_secs)
                self._requested_data_type = "ohlc"
                self._producing_data_type = "trade"
                if "timeframe" in params:
                    self._timeframe = params.get("timeframe", "1Min")

            case DataType.OHLC:
                # - requested restore bars from OHLC
                self._transformer = RestoredBarsFromOHLC(open_close_time_shift_secs=open_close_time_indent_secs)
                self._requested_data_type = "ohlc"
                self._producing_data_type = "ohlc"
                if "timeframe" in params:
                    self._timeframe = params.get("timeframe", "1Min")

            case DataType.TRADE:
                self._requested_data_type = "trade"
                self._producing_data_type = "trade"
                self._transformer = AsTrades()

            case DataType.QUOTE:
                self._requested_data_type = "orderbook"
                self._producing_data_type = "quote"  # ???
                self._transformer = AsQuotes()

            case _:
                self._requested_data_type = subtype
                self._producing_data_type = subtype
                self._transformer = AsDict()

        self._warmup_period = warmup_period
        self._warmed = {}
        self._specs = []
        self._chunksize = chunksize

    @staticmethod
    def _make_request_id(instrument: Instrument) -> str:
        return f"{instrument.exchange}:{instrument.symbol}"

    def attach_instrument(self, instrument: Instrument) -> str:
        _data_id = self._make_request_id(instrument)

        if _data_id not in self._specs:
            self._specs.append(_data_id)
            self._warmed[_data_id] = False

        return self._fetcher_id + "." + _data_id

    def remove_instrument(self, instrument: Instrument) -> str:
        _data_id = self._make_request_id(instrument)

        if _data_id in self._specs:
            self._specs.remove(_data_id)
            del self._warmed[_data_id]

        return self._fetcher_id + "." + _data_id

    def has_instrument(self, instrument: Instrument) -> bool:
        return self._make_request_id(instrument) in self._specs

    def get_instruments_indices(self) -> list[str]:
        return [self._fetcher_id + "." + i for i in self._specs]

    def get_instrument_index(self, instrument: Instrument) -> str:
        return self._fetcher_id + "." + self._make_request_id(instrument)

    def load(
        self, start: str | pd.Timestamp, end: str | pd.Timestamp, to_load: list[Instrument] | None
    ) -> dict[str, Iterator]:
        """
        Loads data for specified instruments within a given time range.

        Parameters:
            - start (str | pd.Timestamp): The start time for data loading, can be a string or a pandas Timestamp.
            - end (str | pd.Timestamp): The end time for data loading, can be a string or a pandas Timestamp.
            - to_load (list[Instrument] | None): A list of instruments to load data for. If None, data for all subscribed instruments is loaded.

        Returns:
            - dict[str, Iterator]: A dictionary where keys are instrument identifiers and values are iterators over the loaded data.
        """
        _requests = self._specs if not to_load else set(self._make_request_id(i) for i in to_load)
        _r_iters = {}

        for _r in _requests:  # - TODO: replace this loop with multi-instrument request after DataReader refactoring
            if _r in self._specs:
                _start = pd.Timestamp(start)
                if self._warmup_period and not self._warmed.get(_r):
                    _start -= self._warmup_period
                    self._warmed[_r] = True

                _args = dict(
                    data_id=_r,
                    start=_start,
                    stop=end,
                    transform=self._transformer,
                    data_type=self._requested_data_type,
                    chunksize=self._chunksize,
                )

                if self._timeframe:
                    _args["timeframe"] = self._timeframe

                try:
                    _r_iters[self._fetcher_id + "." + _r] = self._reader.read(**_args)  # type: ignore
                except Exception as e:
                    logger.error(f">>> (DataFetcher::load) - failed to load <g>'{self._fetcher_id}'</g> data: {e}")
            else:
                raise IndexError(
                    f"Instrument {_r} is not subscribed for this data {self._requested_data_type} in {self._fetcher_id} !"
                )

        return _r_iters

    def __repr__(self) -> str:
        return f"{self._requested_data_type}({self._params}) (-{self._warmup_period if self._warmup_period else '--'}) [{','.join(self._specs)}] :-> {self._transformer.__class__.__name__}"


class IterableSimulationData(Iterator):
    """
    This class is a crucial component for backtesting system.
    It provides a flexible and efficient way to simulate market data feeds for strategy testing.

    Key Features:
        - Supports multiple data types (OHLC, trades, quotes) and instruments.
        - Allows for dynamic addition and removal of instruments during simulation.
        - Handles warmup periods for data preloading.
        - Manages historical and current data distinction during iteration.
        - Utilizes a data slicer (IteratedDataStreamsSlicer) to merge and order data from multiple sources.

    TODO:
        1. think how to provide initial "market quote" for each instrument
        2. optimization for historical data (return bunch of history instead of each bar in next(...))
    """

    _readers: dict[str, DataReader]
    _subtyped_fetchers: dict[str, DataFetcher]
    _warmups: dict[str, pd.Timedelta]
    _instruments: dict[str, tuple[Instrument, DataFetcher, DataType]]
    _open_close_time_indent_secs: int | float

    _slicer_ctrl: IteratedDataStreamsSlicer | None = None
    _slicing_iterator: Iterator | None = None
    _start: pd.Timestamp | None = None
    _stop: pd.Timestamp | None = None
    _current_time: int | None = None

    def __init__(
        self,
        readers: dict[str, DataReader],
        open_close_time_indent_secs=1,  # open/close ticks shift
    ):
        self._readers = dict(readers)
        self._instruments = {}
        self._subtyped_fetchers = {}
        self._warmups = {}
        self._open_close_time_indent_secs = open_close_time_indent_secs

    def set_typed_reader(self, data_type: str, reader: DataReader):
        self._readers[data_type] = reader
        if _fetcher := self._subtyped_fetchers.get(data_type):
            _fetcher._reader = reader

    def set_warmup_period(self, subscription: str, warmup_period: str | None = None):
        if warmup_period:
            _access_key, _, _ = self._parse_subscription_spec(subscription)
            self._warmups[_access_key] = pd.Timedelta(warmup_period)

    def _parse_subscription_spec(self, subscription: str) -> tuple[str, str, dict[str, object]]:
        _subtype, _params = DataType.from_str(subscription)
        match _subtype:
            case DataType.OHLC | DataType.OHLC_QUOTES:
                _timeframe = _params.get("timeframe", "1Min")
                _access_key = f"{_subtype}.{_timeframe}"
            case DataType.TRADE | DataType.QUOTE:
                _access_key = f"{_subtype}"
            case _:
                # - any arbitrary data type is passed as is
                _params = {}
                _subtype = subscription
                _access_key = f"{_subtype}"
        return _access_key, _subtype, _params

    def add_instruments_for_subscription(self, subscription: str, instruments: list[Instrument] | Instrument):
        instruments = instruments if isinstance(instruments, list) else [instruments]
        _subt_key, _data_type, _params = self._parse_subscription_spec(subscription)

        fetcher = self._subtyped_fetchers.get(_subt_key)
        if not fetcher:
            _reader = self._readers.get(_data_type)

            if _reader is None:
                raise SimulationError(f"No reader configured for data type: {_data_type}")

            self._subtyped_fetchers[_subt_key] = (
                fetcher := DataFetcher(
                    _subt_key,
                    _reader,
                    _data_type,
                    _params,
                    warmup_period=self._warmups.get(_subt_key),
                    open_close_time_indent_secs=self._open_close_time_indent_secs,
                )
            )

        _instrs_to_preload = []
        for i in instruments:
            if not fetcher.has_instrument(i):
                idx = fetcher.attach_instrument(i)
                self._instruments[idx] = (i, fetcher, subscription)  # type: ignore
                _instrs_to_preload.append(i)

        if self.is_running and _instrs_to_preload:
            self._slicer_ctrl += fetcher.load(
                pd.Timestamp(self._current_time, unit="ns"),  # type: ignore
                self._stop,  # type: ignore
                _instrs_to_preload,
            )

    def peek_historical_data(self, instrument: Instrument, subscription: str) -> list[Timestamped]:
        """
        Retrieves historical data for a specified instrument and subscription type up to the current simulation time.

        Parameters:
            - instrument (Instrument): instrument for which historical data is requested.
            - subscription (str): type of data subscription (e.g., OHLC, trades, quotes) for the instrument.

        Returns:
            - list[Timestamped]: A list of historical data elements for the specified instrument and subscription type
            that occurred before the current simulation time. If the simulation is not running, returns an empty list.

        Raises:
            SimulationError: If the instrument does not have the specified subscription in the simulation data provider.
        """
        if not self.has_subscription(instrument, subscription):
            raise SimulationError(
                f"Instrument: {instrument} has no subscription: {subscription} in this simulation data provider"
            )

        if not self.is_running:
            return []

        _subt_key, _, _ = self._parse_subscription_spec(subscription)
        _i_key = self._subtyped_fetchers[_subt_key].get_instrument_index(instrument)

        assert self._slicer_ctrl is not None and self._current_time is not None

        # fetch historical data for current time
        return self._slicer_ctrl.fetch_before_time(_i_key, self._current_time)

    def get_instruments_for_subscription(self, subscription: str) -> list[Instrument]:
        if subscription == DataType.ALL:
            return list((i for i, *_ in self._instruments.values()))

        _subt_key, _, _ = self._parse_subscription_spec(subscription)
        if (fetcher := self._subtyped_fetchers.get(_subt_key)) is not None:
            return [self._instruments[k][0] for k in fetcher.get_instruments_indices()]

        return []

    def get_subscriptions_for_instrument(self, instrument: Instrument | None) -> list[str]:
        r = []
        for i, f, s in self._instruments.values():
            if instrument is not None:
                if i == instrument:
                    r.append(s)
            else:
                r.append(s)
        return list(set(r))

    def has_subscription(self, instrument: Instrument, subscription_type: str) -> bool:
        for i, f, s in self._instruments.values():
            if i == instrument and s == subscription_type:
                return True
        return False

    def remove_instruments_from_subscription(self, subscription: str, instruments: list[Instrument] | Instrument):
        def _remove_from_fetcher(_subt_key: str, instruments: list[Instrument]):
            fetcher = self._subtyped_fetchers.get(_subt_key)
            if not fetcher:
                logger.warning(f"No configured data fetcher for '{_subt_key}' subscription !")
                return

            _keys_to_remove = []
            for i in instruments:
                # - try to remove from data fetcher
                if idx := fetcher.remove_instrument(i):
                    if idx in self._instruments:
                        self._instruments.pop(idx)
                        _keys_to_remove.append(idx)

            # print("REMOVING FROM:", _keys_to_remove)
            if self.is_running and _keys_to_remove:
                self._slicer_ctrl.remove(_keys_to_remove)  # type: ignore

        instruments = instruments if isinstance(instruments, list) else [instruments]

        # - if we want to remove instruments from all subscriptions
        if subscription == DataType.ALL:
            _f_keys = list(self._subtyped_fetchers.keys())
            for s in _f_keys:
                _remove_from_fetcher(s, instruments)
            return

        _subt_key, _, _ = self._parse_subscription_spec(subscription)
        _remove_from_fetcher(_subt_key, instruments)

    @property
    def is_running(self) -> bool:
        return self._current_time is not None

    def create_iterable(self, start: str | pd.Timestamp, stop: str | pd.Timestamp) -> Iterator:
        self._start = pd.Timestamp(start)
        self._stop = pd.Timestamp(stop)
        self._current_time = None
        self._slicer_ctrl = IteratedDataStreamsSlicer()
        return self

    def __iter__(self) -> Iterator:
        assert self._start is not None
        self._current_time = int(pd.Timestamp(self._start).timestamp() * 1e9)
        _ct_timestap = pd.Timestamp(self._current_time, unit="ns")

        for f in self._subtyped_fetchers.values():
            logger.debug(
                f"  [<c>IteratedDataStreamsSlicer</c>] :: Preloading initial data for {f._fetcher_id} {self._start} : {self._stop} ..."
            )
            self._slicer_ctrl += f.load(_ct_timestap, self._stop, None)  # type: ignore

        self._slicing_iterator = iter(self._slicer_ctrl)
        return self

    def __next__(self) -> tuple[Instrument, str, Timestamped, bool]:  # type: ignore
        try:
            while data := next(self._slicing_iterator):  # type: ignore
                k, t, v = data
                instr, fetcher, subt = self._instruments[k]
                data_type = fetcher._producing_data_type
                _is_historical = False
                if t < self._current_time:  # type: ignore
                    _is_historical = True
                else:
                    # only update the current time if the event is not historical
                    self._current_time = t

                return instr, data_type, v, _is_historical
        except StopIteration as e:  # noqa: F841
            raise StopIteration



from dataclasses import dataclass
from operator import neg

import numpy as np
from sortedcontainers import SortedDict

from qubx import logger
from qubx.core.basics import (
    OPTION_FILL_AT_SIGNAL_PRICE,
    Deal,
    Instrument,
    ITimeProvider,
    Order,
    OrderSide,
    OrderType,
    TransactionCostsCalculator,
    dt_64,
)
from qubx.core.exceptions import (
    ExchangeError,
    InvalidOrder,
)
from qubx.core.series import Quote


@dataclass
class OmeReport:
    timestamp: dt_64
    order: Order
    exec: Deal | None


class OrdersManagementEngine:
    instrument: Instrument
    time_service: ITimeProvider
    active_orders: dict[str, Order]
    stop_orders: dict[str, Order]
    asks: SortedDict[float, list[str]]
    bids: SortedDict[float, list[str]]
    bbo: Quote | None  # current best bid/ask order book (simplest impl)
    __order_id: int
    __trade_id: int
    _fill_stops_at_price: bool

    def __init__(
        self,
        instrument: Instrument,
        time_provider: ITimeProvider,
        tcc: TransactionCostsCalculator,
        fill_stop_order_at_price: bool = False,  # emulate stop orders execution at order's exact limit price
        debug: bool = True,
    ) -> None:
        self.instrument = instrument
        self.time_service = time_provider
        self.tcc = tcc
        self.asks = SortedDict()
        self.bids = SortedDict(neg)
        self.active_orders = dict()
        self.stop_orders = dict()
        self.bbo = None
        self.__order_id = 100000
        self.__trade_id = 100000
        self._fill_stops_at_price = fill_stop_order_at_price
        if not debug:
            self._dbg = lambda message, **kwargs: None

    def _generate_order_id(self) -> str:
        self.__order_id += 1
        return "SIM-ORDER-" + self.instrument.symbol + "-" + str(self.__order_id)

    def _generate_trade_id(self) -> str:
        self.__trade_id += 1
        return "SIM-EXEC-" + self.instrument.symbol + "-" + str(self.__trade_id)

    def get_quote(self) -> Quote:
        return self.bbo

    def get_open_orders(self) -> list[Order]:
        return list(self.active_orders.values()) + list(self.stop_orders.values())

    def update_bbo(self, quote: Quote) -> list[OmeReport]:
        timestamp = self.time_service.time()
        rep = []

        if self.bbo is not None:
            if quote.bid >= self.bbo.ask:
                for level in self.asks.irange(0, quote.bid):
                    for order_id in self.asks[level]:
                        order = self.active_orders.pop(order_id)
                        rep.append(self._execute_order(timestamp, order.price, order, False))
                    self.asks.pop(level)

            if quote.ask <= self.bbo.bid:
                for level in self.bids.irange(np.inf, quote.ask):
                    for order_id in self.bids[level]:
                        order = self.active_orders.pop(order_id)
                        rep.append(self._execute_order(timestamp, order.price, order, False))
                    self.bids.pop(level)

            # - processing stop orders
            for soid in list(self.stop_orders.keys()):
                so = self.stop_orders[soid]
                _emulate_price_exec = self._fill_stops_at_price or so.options.get(OPTION_FILL_AT_SIGNAL_PRICE, False)

                if so.side == "BUY" and quote.ask >= so.price:
                    _exec_price = quote.ask if not _emulate_price_exec else so.price
                    self.stop_orders.pop(soid)
                    rep.append(self._execute_order(timestamp, _exec_price, so, True))
                elif so.side == "SELL" and quote.bid <= so.price:
                    _exec_price = quote.bid if not _emulate_price_exec else so.price
                    self.stop_orders.pop(soid)
                    rep.append(self._execute_order(timestamp, _exec_price, so, True))

        self.bbo = quote
        return rep

    def place_order(
        self,
        order_side: OrderSide,
        order_type: OrderType,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **options,
    ) -> OmeReport:
        if self.bbo is None:
            raise ExchangeError(f"Simulator is not ready for order management - no quote for {self.instrument.symbol}")

        # - validate order parameters
        self._validate_order(order_side, order_type, amount, price, time_in_force)

        timestamp = self.time_service.time()
        order = Order(
            self._generate_order_id(),
            order_type,
            self.instrument,
            timestamp,
            amount,
            price if price is not None else 0,
            order_side,
            "NEW",
            time_in_force,
            client_id,
            options=options,
        )

        return self._process_order(timestamp, order)

    def _dbg(self, message, **kwargs) -> None:
        logger.debug(f"    [<y>OME</y>(<g>{self.instrument}</g>)] :: {message}", **kwargs)

    def _process_order(self, timestamp: dt_64, order: Order) -> OmeReport:
        if order.status in ["CLOSED", "CANCELED"]:
            raise InvalidOrder(f"Order {order.id} is already closed or canceled.")

        buy_side = order.side == "BUY"
        c_ask = self.bbo.ask
        c_bid = self.bbo.bid

        # - check if order can be "executed" immediately
        exec_price = None
        _need_update_book = False

        if order.type == "MARKET":
            exec_price = c_ask if buy_side else c_bid

        elif order.type == "LIMIT":
            _need_update_book = True
            if (buy_side and order.price >= c_ask) or (not buy_side and order.price <= c_bid):
                exec_price = c_ask if buy_side else c_bid

        elif order.type == "STOP_MARKET":
            # - it processes stop orders separately without adding to orderbook (as on real exchanges)
            order.status = "OPEN"
            self.stop_orders[order.id] = order

        elif order.type == "STOP_LIMIT":
            # TODO: (OME) check trigger conditions in options etc
            raise NotImplementedError("'STOP_LIMIT' order is not supported in Qubx simulator yet !")

        # - if order must be "executed" immediately
        if exec_price is not None:
            return self._execute_order(timestamp, exec_price, order, True)

        # - processing limit orders
        if _need_update_book:
            if buy_side:
                self.bids.setdefault(order.price, list()).append(order.id)
            else:
                self.asks.setdefault(order.price, list()).append(order.id)

            order.status = "OPEN"
            self.active_orders[order.id] = order

        self._dbg(f"registered {order.id} {order.type} {order.side} {order.quantity} {order.price}")
        return OmeReport(timestamp, order, None)

    def _execute_order(self, timestamp: dt_64, exec_price: float, order: Order, taker: bool) -> OmeReport:
        order.status = "CLOSED"
        self._dbg(f"<red>{order.id}</red> {order.type} {order.side} {order.quantity} executed at {exec_price}")
        return OmeReport(
            timestamp,
            order,
            Deal(
                id=self._generate_trade_id(),
                order_id=order.id,
                time=timestamp,
                amount=order.quantity if order.side == "BUY" else -order.quantity,
                price=exec_price,
                aggressive=taker,
                fee_amount=self.tcc.get_execution_fees(
                    instrument=self.instrument, exec_price=exec_price, amount=order.quantity, crossed_market=taker
                ),
                fee_currency=self.instrument.quote,
            ),
        )

    def _validate_order(
        self, order_side: str, order_type: str, amount: float, price: float | None, time_in_force: str
    ) -> None:
        if order_side.upper() not in ["BUY", "SELL"]:
            raise InvalidOrder("Invalid order side. Only BUY or SELL is allowed.")

        _ot = order_type.upper()
        if _ot not in ["LIMIT", "MARKET", "STOP_MARKET", "STOP_LIMIT"]:
            raise InvalidOrder("Invalid order type. Only LIMIT, MARKET, STOP_MARKET, STOP_LIMIT are supported.")

        if amount <= 0:
            raise InvalidOrder("Invalid order amount. Amount must be positive.")

        if (_ot == "LIMIT" or _ot.startswith("STOP")) and (price is None or price <= 0):
            raise InvalidOrder("Invalid order price. Price must be positively defined for LIMIT or STOP orders.")

        if time_in_force.upper() not in ["GTC", "IOC"]:
            raise InvalidOrder("Invalid time in force. Only GTC or IOC is supported for now.")

        if _ot.startswith("STOP"):
            assert price is not None
            c_ask, c_bid = self.bbo.ask, self.bbo.bid
            if (order_side == "BUY" and c_ask >= price) or (order_side == "SELL" and c_bid <= price):
                raise ExchangeError(
                    f"Stop price would trigger immediately: STOP_MARKET {order_side} {amount} of {self.instrument.symbol} at {price} | market: {c_ask} / {c_bid}"
                )

    def cancel_order(self, order_id: str) -> OmeReport:
        # - check limit orders
        if order_id in self.active_orders:
            order = self.active_orders.pop(order_id)
            if order.side == "BUY":
                oids = self.bids[order.price]
                oids.remove(order_id)
                if not oids:
                    self.bids.pop(order.price)
            else:
                oids = self.asks[order.price]
                oids.remove(order_id)
                if not oids:
                    self.asks.pop(order.price)
        # - check stop orders
        elif order_id in self.stop_orders:
            order = self.stop_orders.pop(order_id)
        # - wrong order_id
        else:
            raise InvalidOrder(f"Order {order_id} not found for {self.instrument.symbol}")

        order.status = "CANCELED"
        self._dbg(f"{order.id} {order.type} {order.side} {order.quantity} canceled")
        return OmeReport(self.time_service.time(), order, None)

    def __str__(self) -> str:
        _a, _b = True, True

        timestamp = self.time_service.time()
        _s = f"= = ({np.datetime64(timestamp, 'ns')}) = =\n"
        for k, v in reversed(self.asks.items()):
            _sizes = ",".join([f"{self.active_orders[o].quantity}" for o in v])
            _s += f"  {k} : [{_sizes}]\n"
            if k == self.bbo.ask:
                _a = False

        if _a:
            _s += f"  {self.bbo.ask} : \n"
        _s += "- - - - - - - - - - - - - - - - - - - -\n"

        _s1 = ""
        for k, v in self.bids.items():
            _sizes = ",".join([f"{self.active_orders[o].quantity}" for o in v])
            _s1 += f"  {k} : [{_sizes}]\n"
            if k == self.bbo.bid:
                _b = False
        _s1 += "= = = = = = = = = = = = = = = = = = = =\n"

        _s1 = f"  {self.bbo.bid} : \n" + _s1 if _b else _s1

        return _s + _s1



from typing import Any

import pandas as pd

from qubx.core.basics import Instrument, ITimeProvider, dt_64
from qubx.core.helpers import CachedMarketDataHolder
from qubx.core.interfaces import (
    IDataProvider,
    IMarketManager,
    IUniverseManager,
)
from qubx.core.lookups import lookup
from qubx.core.series import OHLCV, Quote
from qubx.data.readers import DataReader
from qubx.utils import convert_seconds_to_str


class MarketManager(IMarketManager):
    _time_provider: ITimeProvider
    _cache: CachedMarketDataHolder
    _data_provider: IDataProvider
    _universe_manager: IUniverseManager
    _aux_data_provider: DataReader | None

    def __init__(
        self,
        time_provider: ITimeProvider,
        cache: CachedMarketDataHolder,
        data_provider: IDataProvider,
        universe_manager: IUniverseManager,
        aux_data_provider: DataReader | None = None,
    ):
        self._time_provider = time_provider
        self._cache = cache
        self._data_provider = data_provider
        self._universe_manager = universe_manager
        self._aux_data_provider = aux_data_provider

    def time(self) -> dt_64:
        return self._time_provider.time()

    def ohlc(
        self,
        instrument: Instrument,
        timeframe: str | None = None,
        length: int | None = None,
    ) -> OHLCV:
        timeframe = timeframe or convert_seconds_to_str(
            int(pd.Timedelta(self._cache.default_timeframe).total_seconds())
        )
        rc = self._cache.get_ohlcv(instrument, timeframe)

        # - check if we need to fetch more data
        _need_history_request = False
        if (_l_rc := len(rc)) > 0:
            _last_bar_time = rc[0].time
            _timeframe_ns = pd.Timedelta(timeframe).asm8.item()

            # - check if we need to fetch more data
            if (_last_bar_time + _timeframe_ns < self._data_provider.time_provider.time().item()) or (
                length and _l_rc < length
            ):
                _need_history_request = True

        else:
            _need_history_request = True

        # - send request for historical data
        if _need_history_request and length is not None:
            bars = self._data_provider.get_ohlc(instrument, timeframe, length)
            rc = self._cache.update_by_bars(instrument, timeframe, bars)
        return rc

    def quote(self, instrument: Instrument) -> Quote | None:
        return self._data_provider.get_quote(instrument)

    def get_data(self, instrument: Instrument, sub_type: str) -> list[Any]:
        return self._cache.get_data(instrument, sub_type)

    def get_aux_data(self, data_id: str, **parameters) -> pd.DataFrame | None:
        return self._aux_data_provider.get_aux_data(data_id, **parameters) if self._aux_data_provider else None

    def get_instruments(self) -> list[Instrument]:
        return self._universe_manager.instruments

    def query_instrument(self, symbol: str, exchange: str) -> Instrument | None:
        return lookup.find_symbol(exchange, symbol)

    def exchanges(self) -> list[str]:
        """
        What exchanges are supported by the market manager.
        Theoretically it can manage multiple exchanges.
        """
        return [self._data_provider.exchange()]



from dataclasses import field
from os.path import exists, expanduser
from pathlib import Path
from typing import Any, Iterable

import pandas as pd
from pyarrow import csv

from qubx.utils.time import handle_start_stop, infer_series_frequency

from .readers import CsvStorageDataReader, DataReader, DataTransformer, _recognize_t

TARDIS_EXCHANGE_MAPPERS = {
    "bitfinex.f": "bitfinex-derivatives",
    "binance.um": "binance-futures",
}


class TardisCsvDataReader(DataReader):
    def __init__(self, path: str | Path) -> None:
        _path = expanduser(path)
        if not exists(_path):
            raise ValueError(f"Folder is not found at {path}")
        self.path = Path(_path)

    def get_names(self, exchange: str | None = None, data_type: str | None = None) -> list[str]:
        symbols = []
        exchanges = [exchange] if exchange else self.get_exchanges()
        for exchange in exchanges:
            exchange_path = Path(self.path) / exchange
            if not exists(exchange_path):
                raise ValueError(f"Exchange is not found at {exchange_path}")
            data_types = [data_type] if data_type else self.get_data_types(exchange)
            for data_type in data_types:
                data_type_path = exchange_path / data_type
                if not exists(data_type_path):
                    return []
                symbols += self._get_symbols(data_type_path)
        return symbols

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timeframe=None,
        data_type="trades",
    ) -> Iterable | Any:
        if chunksize > 0:
            raise NotImplementedError("Chunksize is not supported for TardisCsvDataReader")
        exchange, symbol = data_id.split(":")
        _exchange = exchange.lower()
        _exchange = TARDIS_EXCHANGE_MAPPERS.get(_exchange, _exchange)
        t_0, t_1 = handle_start_stop(start, stop, lambda x: pd.Timestamp(x).date().isoformat())
        _path = self.path / _exchange / data_type
        if not _path.exists():
            raise ValueError(f"Data type is not found at {_path}")
        _files = sorted(_path.glob(f"*_{symbol}.csv.gz"))
        if not _files:
            return None
        _dates = [file.stem.split("_")[0] for file in _files]
        if t_0 is None:
            t_0 = _dates[0]
        if t_1 is None:
            t_1 = _dates[-1]
        _filt_files = [file for file in _files if t_0 <= file.stem.split("_")[0] <= t_1]

        tables = []
        fieldnames = None
        for f_path in _filt_files:
            table = csv.read_csv(
                f_path,
                parse_options=csv.ParseOptions(ignore_empty_lines=True),
            )
            if not fieldnames:
                fieldnames = table.column_names
            tables.append(table.to_pandas())

        transform.start_transform(data_id, fieldnames, start=start, stop=stop)
        raw_data = pd.concat(tables).to_numpy()
        transform.process_data(raw_data)

        return transform.collect()

    def get_exchanges(self) -> list[str]:
        return [exchange.name for exchange in self.path.iterdir() if exchange.is_dir()]

    def get_data_types(self, exchange: str) -> list[str]:
        exchange_path = Path(self.path) / exchange
        return [data_type.name for data_type in exchange_path.iterdir() if data_type.is_dir()]

    def _get_symbols(self, data_type_path: Path) -> list[str]:
        symbols = set()
        for file in data_type_path.glob("*.gz"):
            parts = file.stem.replace(".csv", "").split("_")
            if len(parts) == 2:
                symbols.add(parts[1])
        return list(symbols)



import itertools
import os
import re
from functools import wraps
from os.path import exists, join
from typing import Any, Iterable, Iterator, List, Set, Union

import numpy as np
import pandas as pd
import psycopg as pg
import pyarrow as pa
from pyarrow import csv, table

from qubx import logger
from qubx.core.basics import DataType, TimestampedDict
from qubx.core.series import OHLCV, Bar, Quote, Trade
from qubx.pandaz.utils import ohlc_resample, srows
from qubx.utils.time import handle_start_stop, infer_series_frequency


def convert_timedelta_to_numpy(x: str) -> int:
    return pd.Timedelta(x).to_numpy().item()


D1, H1 = convert_timedelta_to_numpy("1D"), convert_timedelta_to_numpy("1h")
MS1 = 1_000_000
S1 = 1000 * MS1
M1 = 60 * S1

DEFAULT_DAILY_SESSION = (convert_timedelta_to_numpy("00:00:00.100"), convert_timedelta_to_numpy("23:59:59.900"))
STOCK_DAILY_SESSION = (convert_timedelta_to_numpy("9:30:00.100"), convert_timedelta_to_numpy("15:59:59.900"))
CME_FUTURES_DAILY_SESSION = (convert_timedelta_to_numpy("8:30:00.100"), convert_timedelta_to_numpy("15:14:59.900"))


def _recognize_t(t: Union[int, str], defaultvalue, timeunit) -> int:
    if isinstance(t, (str, pd.Timestamp)):
        try:
            return np.datetime64(t, timeunit)
        except (ValueError, TypeError) as e:
            logger.debug(f"Failed to convert time {t} to datetime64: {e}")
    return defaultvalue


def _time(t, timestamp_units: str) -> int:
    t = int(t) if isinstance(t, float) else t
    if timestamp_units == "ns":
        return np.datetime64(t, "ns").item()
    return np.datetime64(t, timestamp_units).astype("datetime64[ns]").item()


def _find_column_index_in_list(xs, *args):
    xs = [x.lower() for x in xs]
    for a in args:
        ai = a.lower()
        if ai in xs:
            return xs.index(ai)
    raise IndexError(f"Can't find any specified columns from [{args}] in provided list: {xs}")


def _list_to_chunked_iterator(data: list[Any], chunksize: int) -> Iterable:
    it = iter(data)
    chunk = list(itertools.islice(it, chunksize))
    while chunk:
        yield chunk
        chunk = list(itertools.islice(it, chunksize))


def _find_time_col_idx(column_names):
    return _find_column_index_in_list(column_names, "time", "timestamp", "datetime", "date", "open_time", "ts")


class DataTransformer:
    def __init__(self) -> None:
        self.buffer = []
        self._column_names = []

    def start_transform(
        self,
        name: str,
        column_names: List[str],
        start: str | None = None,
        stop: str | None = None,
    ):
        self._column_names = column_names
        self.buffer = []

    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None:
            self.buffer.extend(rows_data)

    def collect(self) -> Any:
        return self.buffer


class DataReader:
    def get_names(self, **kwargs) -> List[str]:
        """
        TODO: not sure we really need this !
        """
        raise NotImplementedError("get_names() method is not implemented")

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        **kwargs,
    ) -> Iterator | List:
        raise NotImplementedError("read() method is not implemented")

    def get_aux_data_ids(self) -> Set[str]:
        """
        Returns list of all auxiliary data IDs available for this data reader
        """

        def _list_methods(cls):
            _meth = []
            for k, s in cls.__dict__.items():
                if (
                    k.startswith("get_")
                    and k not in ["get_names", "get_symbols", "get_time_ranges", "get_aux_data_ids", "get_aux_data"]
                    and callable(s)
                ):
                    _meth.append(k[4:])
            return _meth

        _d_ids = _list_methods(self.__class__)
        for bc in self.__class__.__bases__:
            _d_ids.extend(_list_methods(bc))
        return set(_d_ids)

    def get_aux_data(self, data_id: str, **kwargs) -> Any:
        """
        Returns auxiliary data for the specified data ID
        """
        if hasattr(self, f"get_{data_id}"):
            return getattr(self, f"get_{data_id}")(**kwargs)
        raise ValueError(
            f"{self.__class__.__name__} doesn't have getter for '{data_id}' auxiliary data. Available data: {self.get_aux_data_ids()}"
        )

    def get_symbols(self, exchange: str, dtype: str) -> list[str]:
        raise NotImplementedError("get_symbols() method is not implemented")

    def get_time_ranges(self, symbol: str, dtype: str) -> tuple[np.datetime64, np.datetime64]:
        """
        Returns first and last time for the specified symbol and data type in the reader's storage
        """
        raise NotImplementedError("get_time_ranges() method is not implemented")


class CsvStorageDataReader(DataReader):
    """
    Data reader for timeseries data stored as csv files in the specified directory
    """

    def __init__(self, path: str) -> None:
        _path = os.path.expanduser(path)
        if not exists(_path):
            raise ValueError(f"Folder is not found at {path}")
        self.path = _path

    def __find_time_idx(self, arr: pa.ChunkedArray, v) -> int:
        ix = arr.index(v).as_py()
        if ix < 0:
            for c in arr.iterchunks():
                a = c.to_numpy()
                ix = np.searchsorted(a, v, side="right")
                if ix > 0 and ix < len(c):
                    ix = arr.index(a[ix]).as_py() - 1
                    break
        return ix

    def __check_file_name(self, name: str) -> str | None:
        _f = join(self.path, name.replace(":", os.sep))
        for sfx in [".csv", ".csv.gz", ""]:
            if exists(p := (_f + sfx)):
                return p
        return None

    def __try_read_data(
        self, data_id: str, start: str | None = None, stop: str | None = None, timestamp_formatters=None
    ) -> tuple[table, np.ndarray, Any, list[str], int, int]:
        f_path = self.__check_file_name(data_id)
        if not f_path:
            ValueError(f"Can't find any csv data for {data_id} in {self.path} !")

        convert_options = None
        if timestamp_formatters is not None:
            convert_options = csv.ConvertOptions(timestamp_parsers=timestamp_formatters)

        table = csv.read_csv(
            f_path,
            parse_options=csv.ParseOptions(ignore_empty_lines=True),
            convert_options=convert_options,
        )
        fieldnames = table.column_names

        # - try to find range to load
        start_idx, stop_idx = 0, table.num_rows
        try:
            _time_field_idx = _find_time_col_idx(fieldnames)
            _time_type = table.field(_time_field_idx).type
            _time_unit = _time_type.unit if hasattr(_time_type, "unit") else "ms"
            _time_data = table[_time_field_idx]

            # - check if need convert time to primitive types (i.e. Date32 -> timestamp[x])
            _time_cast_function = lambda xs: xs
            if _time_type != pa.timestamp(_time_unit):
                _time_cast_function = lambda xs: xs.cast(pa.timestamp(_time_unit))
                _time_data = _time_cast_function(_time_data)

            # - preprocessing start and stop
            t_0, t_1 = handle_start_stop(start, stop, convert=lambda x: _recognize_t(x, None, _time_unit))

            # - check requested range
            if t_0:
                start_idx = self.__find_time_idx(_time_data, t_0)
                if start_idx >= table.num_rows:
                    # - no data for requested start date
                    return table, _time_data, _time_unit, fieldnames, -1, -1

            if t_1:
                stop_idx = self.__find_time_idx(_time_data, t_1)
                if stop_idx < 0 or stop_idx < start_idx:
                    stop_idx = table.num_rows

        except Exception as exc:
            logger.warning(f"exception [{exc}] during preprocessing '{f_path}'")

        return table, _time_data, _time_unit, fieldnames, start_idx, stop_idx

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timestamp_formatters=None,
        timeframe=None,
        **kwargs,
    ) -> Iterable | Any:
        table, _, _, fieldnames, start_idx, stop_idx = self.__try_read_data(data_id, start, stop, timestamp_formatters)
        if start_idx < 0 or stop_idx < 0:
            return None
        length = stop_idx - start_idx + 1
        selected_table = table.slice(start_idx, length)

        # - in this case we want to return iterable chunks of data
        if chunksize > 0:

            def _iter_chunks():
                for n in range(0, length // chunksize + 1):
                    transform.start_transform(data_id, fieldnames, start=start, stop=stop)
                    raw_data = selected_table[n * chunksize : min((n + 1) * chunksize, length)].to_pandas().to_numpy()
                    transform.process_data(raw_data)
                    yield transform.collect()

            return _iter_chunks()

        transform.start_transform(data_id, fieldnames, start=start, stop=stop)
        raw_data = selected_table.to_pandas().to_numpy()
        transform.process_data(raw_data)
        return transform.collect()

    def get_candles(
        self,
        exchange: str,
        symbols: list[str],
        start: str | pd.Timestamp,
        stop: str | pd.Timestamp,
        timeframe: str | None = None,
    ) -> pd.DataFrame:
        """
        Returns pandas DataFrame of candles for given exchange and symbols within specified time range and timeframe
        """
        _r = []
        for symbol in symbols:
            x = self.read(
                f"{exchange}:{symbol}", start=start, stop=stop, timeframe=timeframe, transform=AsPandasFrame()
            )
            if x is not None:
                if timeframe is not None:
                    x = ohlc_resample(x, timeframe)
                _r.append(x.assign(symbol=symbol.upper(), timestamp=x.index))  # type: ignore
        return srows(*_r).set_index(["timestamp", "symbol"]) if _r else pd.DataFrame()

    def get_names(self, **kwargs) -> List[str]:
        _n = []
        for root, _, files in os.walk(self.path):
            path = root.split(os.sep)
            for file in files:
                if re.match(r"(.*)\.csv(.gz)?$", file):
                    f = path[-1]
                    n = file.split(".")[0]
                    if f == self.path:
                        name = n
                    else:
                        name = f"{f}:{n}" if f else n
                    _n.append(name)
        return _n

    def get_symbols(self, exchange: str, dtype: str) -> list[str]:
        return self.get_names()

    def get_time_ranges(self, symbol: str, dtype: str) -> tuple[np.datetime64, np.datetime64]:
        _, _time_data, _time_unit, _, start_idx, stop_idx = self.__try_read_data(symbol, None, None, None)
        return (
            np.datetime64(_time_data[start_idx].value, _time_unit),
            np.datetime64(_time_data[stop_idx - 1].value, _time_unit),
        )


class InMemoryDataFrameReader(DataReader):
    """
    Data reader for pandas DataFrames
    """

    exchange: str | None
    _data: dict[str, pd.DataFrame | pd.Series]

    def __init__(self, data: dict[str, pd.DataFrame | pd.Series], exchange: str | None = None) -> None:
        if not isinstance(data, dict):
            raise ValueError("data must be a dictionary of pandas DataFrames")
        self._data = data
        self.exchange = exchange

    def get_names(self, **kwargs) -> list[str]:
        keys = list(self._data.keys())
        if self.exchange:
            return [f"{self.exchange}:{k}" for k in keys]
        return keys

    def _get_data_by_key(self, data_id: str) -> tuple[str, pd.DataFrame | pd.Series]:
        if data_id not in self._data:
            if self.exchange and data_id.startswith(self.exchange):
                data_id = data_id.split(":")[1]
        if (d := self._data.get(data_id)) is None:
            raise ValueError(f"No data found for {data_id}")
        return data_id, d

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        **kwargs,
    ) -> Iterable | list:
        """
        Read and transform data for a given data_id within a specified time range.

        Parameters:
        -----------
        data_id : str
            The identifier for the data to be read.
        start : str | None, optional
            The start time for the data range (inclusive). If None, start from the earliest available data.
        stop : str | None, optional
            The stop time for the data range (inclusive). If None, include data up to the latest available.
        transform : DataTransformer, optional
            An instance of DataTransformer to process the retrieved data. Defaults to DataTransformer().
        chunksize : int, optional
            The size of data chunks to process at a time. If 0, process all data at once. Defaults to 0.
        **kwargs : dict
            Additional keyword arguments for future extensions.

        Returns:
        --------
        Iterable | List
            The processed and transformed data, either as an iterable (if chunksize > 0) or as a list.

        Raises:
        -------
        ValueError
            If no data is found for the given data_id.
        """
        start, stop = handle_start_stop(start, stop)
        data_id, _stored_data = self._get_data_by_key(data_id)

        _sliced_data = _stored_data.loc[start:stop].copy()
        if _tf := kwargs.get("timeframe"):
            _sliced_data = ohlc_resample(_sliced_data, _tf)
            assert isinstance(_sliced_data, pd.DataFrame), "Resampled data should be a DataFrame"
        _sliced_data = _sliced_data.reset_index()

        def _do_transform(values: Iterable, columns: list[str]) -> Iterable:
            transform.start_transform(data_id, columns, start=start, stop=stop)
            transform.process_data(values)
            return transform.collect()

        if chunksize > 0:
            # returns chunked frames
            def _chunked_dataframe(data: np.ndarray, columns: list[str], chunksize: int) -> Iterable:
                it = iter(data)
                chunk = list(itertools.islice(it, chunksize))
                while chunk:
                    yield _do_transform(chunk, columns)
                    chunk = list(itertools.islice(it, chunksize))

            return _chunked_dataframe(_sliced_data.values, list(_sliced_data.columns), chunksize)

        return _do_transform(_sliced_data.values, list(_sliced_data.columns))

    def get_symbols(self, exchange: str, dtype: str) -> list[str]:
        return self.get_names()

    def get_time_ranges(self, symbol: str, dtype: DataType) -> tuple[np.datetime64 | None, np.datetime64 | None]:
        try:
            _, _stored_data = self._get_data_by_key(symbol)
            return _stored_data.index[0], _stored_data.index[-1]
        except ValueError:
            return None, None


class AsPandasFrame(DataTransformer):
    """
    List of records to pandas dataframe transformer
    """

    def __init__(self, timestamp_units=None) -> None:
        self.timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self._time_idx = _find_time_col_idx(column_names)
        self._column_names = column_names
        self._frame = pd.DataFrame()

    def process_data(self, rows_data: Iterable) -> Any:
        self._frame
        p = pd.DataFrame.from_records(rows_data, columns=self._column_names)
        p.set_index(self._column_names[self._time_idx], drop=True, inplace=True)
        p.index = pd.to_datetime(p.index, unit=self.timestamp_units) if self.timestamp_units else p.index
        p.index.rename("timestamp", inplace=True)
        p.sort_index(inplace=True)
        self._frame = pd.concat((self._frame, p), axis=0, sort=True)
        return p

    def collect(self) -> Any:
        return self._frame


class AsOhlcvSeries(DataTransformer):
    """
    Convert incoming data into OHLCV series.

    Incoming data may have one of the following structures:

        ```
        ohlcv:        time,open,high,low,close,volume|quote_volume,(buy_volume)
        quotes:       time,bid,ask,bidsize,asksize
        trades (TAS): time,price,size,(is_taker)
        ```
    """

    timeframe: str | None
    _series: OHLCV | None
    _data_type: str | None

    def __init__(self, timeframe: str | None = None, timestamp_units="ns") -> None:
        super().__init__()
        self.timeframe = timeframe
        self._series = None
        self._data_type = None
        self.timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self._time_idx = _find_time_col_idx(column_names)
        self._volume_idx = None
        self._b_volume_idx = None
        try:
            self._close_idx = _find_column_index_in_list(column_names, "close")
            self._open_idx = _find_column_index_in_list(column_names, "open")
            self._high_idx = _find_column_index_in_list(column_names, "high")
            self._low_idx = _find_column_index_in_list(column_names, "low")

            try:
                self._volume_idx = _find_column_index_in_list(column_names, "quote_volume", "volume", "vol")
            except:
                pass

            try:
                self._b_volume_idx = _find_column_index_in_list(
                    column_names,
                    "taker_buy_volume",
                    "taker_buy_quote_volume",
                    "buy_volume",
                )
            except:
                pass

            self._data_type = "ohlc"
        except:
            try:
                self._ask_idx = _find_column_index_in_list(column_names, "ask")
                self._bid_idx = _find_column_index_in_list(column_names, "bid")
                self._data_type = "quotes"
            except:
                try:
                    self._price_idx = _find_column_index_in_list(column_names, "price")
                    self._size_idx = _find_column_index_in_list(
                        column_names, "quote_qty", "qty", "size", "amount", "volume"
                    )
                    self._taker_idx = None
                    try:
                        self._taker_idx = _find_column_index_in_list(
                            column_names,
                            "is_buyer_maker",
                            "side",
                            "aggressive",
                            "taker",
                            "is_taker",
                        )
                    except:
                        pass

                    self._data_type = "trades"
                except:
                    raise ValueError(f"Can't recognize data for update from header: {column_names}")

        self._column_names = column_names
        self._name = name
        if self.timeframe:
            self._series = OHLCV(self._name, self.timeframe)

    def _proc_ohlc(self, rows_data: List[List]):
        for d in rows_data:
            self._series.update_by_bar(
                _time(d[self._time_idx], self.timestamp_units),
                d[self._open_idx],
                d[self._high_idx],
                d[self._low_idx],
                d[self._close_idx],
                d[self._volume_idx] if self._volume_idx else 0,
                d[self._b_volume_idx] if self._b_volume_idx else 0,
            )

    def _proc_quotes(self, rows_data: List[List]):
        for d in rows_data:
            self._series.update(
                _time(d[self._time_idx], self.timestamp_units),
                (d[self._ask_idx] + d[self._bid_idx]) / 2,
            )

    def _proc_trades(self, rows_data: List[List]):
        for d in rows_data:
            a = d[self._taker_idx] if self._taker_idx else 0
            s = d[self._size_idx]
            b = s if a else 0
            self._series.update(_time(d[self._time_idx], self.timestamp_units), d[self._price_idx], s, b)

    def process_data(self, rows_data: List[List]) -> Any:
        if self._series is None:
            ts = [t[self._time_idx] for t in rows_data[:100]]
            self.timeframe = pd.Timedelta(infer_series_frequency(ts)).asm8.item()

            # - create instance after first data received if
            self._series = OHLCV(self._name, self.timeframe)

        match self._data_type:
            case "ohlc":
                self._proc_ohlc(rows_data)
            case "quotes":
                self._proc_quotes(rows_data)
            case "trades":
                self._proc_trades(rows_data)

        return None

    def collect(self) -> Any:
        return self._series


class AsBars(AsOhlcvSeries):
    """
    Convert incoming data into Bars sequence.

    Incoming data may have one of the following structures:

        ```
        ohlcv:        time,open,high,low,close,volume|quote_volume,(buy_volume)
        quotes:       time,bid,ask,bidsize,asksize
        trades (TAS): time,price,size,(is_taker)
        ```
    """

    def collect(self) -> Any:
        return self._series[::-1] if self._series is not None else None


class AsQuotes(DataTransformer):
    """
    Tries to convert incoming data to list of Quote's
    Data must have appropriate structure: bid, ask, bidsize, asksize and time
    """

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer = list()
        self._time_idx = _find_time_col_idx(column_names)
        self._bid_idx = _find_column_index_in_list(column_names, "bid")
        self._ask_idx = _find_column_index_in_list(column_names, "ask")
        self._bidvol_idx = _find_column_index_in_list(column_names, "bidvol", "bid_vol", "bidsize", "bid_size")
        self._askvol_idx = _find_column_index_in_list(column_names, "askvol", "ask_vol", "asksize", "ask_size")

    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None:
            for d in rows_data:
                t = d[self._time_idx]
                b = d[self._bid_idx]
                a = d[self._ask_idx]
                bv = d[self._bidvol_idx]
                av = d[self._askvol_idx]
                self.buffer.append(Quote(_time(t, "ns"), b, a, bv, av))


class AsTrades(DataTransformer):
    """
    Tries to convert incoming data to list of Trades
    Data must have appropriate structure: price, size, market_maker (optional).
    Market maker column specifies if buyer is a maker or taker.
    """

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer: list[Trade] = list()
        self._time_idx = _find_time_col_idx(column_names)
        self._price_idx = _find_column_index_in_list(column_names, "price")
        self._size_idx = _find_column_index_in_list(column_names, "size")
        try:
            self._side_idx = _find_column_index_in_list(column_names, "market_maker")
        except:
            self._side_idx = None

    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None:
            for d in rows_data:
                t = d[self._time_idx]
                price = d[self._price_idx]
                size = d[self._size_idx]
                side = d[self._side_idx] if self._side_idx else -1
                self.buffer.append(Trade(_time(t, "ns"), price, size, side))


class AsTimestampedRecords(DataTransformer):
    """
    Convert incoming data to list or dictionaries with preprocessed timestamps ('timestamp_ns' and 'timestamp')
    ```
    [
        {
            'open_time': 1711944240000.0,
            'open': 203.219,
            'high': 203.33,
            'low': 203.134,
            'close': 203.175,
            'volume': 10060.0,
            ....
            'timestamp_ns': 1711944240000000000,
            'timestamp': Timestamp('2024-04-01 04:04:00')
        },
        ...
    ] ```
    """

    def __init__(self, timestamp_units: str | None = None) -> None:
        self.timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer = list()
        self._time_idx = _find_time_col_idx(column_names)
        self._column_names = column_names

    def process_data(self, rows_data: Iterable) -> Any:
        self.buffer.extend(rows_data)

    def collect(self) -> Any:
        res = []
        for r in self.buffer:
            t = r[self._time_idx]
            if self.timestamp_units:
                t = _time(t, self.timestamp_units)
            di = dict(zip(self._column_names, r)) | {
                "timestamp_ns": t,
                "timestamp": pd.Timestamp(t),
            }
            res.append(di)
        return res


class RestoredEmulatorHelper(DataTransformer):
    _freq: np.timedelta64 | None = None
    _t_start: int
    _t_mid1: int
    _t_mid2: int
    _t_end: int
    _open_close_time_shift_secs: int

    def __init__(self, daily_session_start_end: tuple, timestamp_units: str, open_close_time_shift_secs: int):
        super().__init__()
        self._d_session_start = daily_session_start_end[0]
        self._d_session_end = daily_session_start_end[1]
        self._timestamp_units = timestamp_units
        self._open_close_time_shift_secs = open_close_time_shift_secs  # type: ignore

    def _detect_emulation_timestamps(self, rows_data: list[list]):
        if self._freq is None:
            ts = [t[self._time_idx] for t in rows_data]
            try:
                self._freq = infer_series_frequency(ts)
            except ValueError:
                logger.warning("Can't determine frequency of incoming data")
                return

            # - timestamps when we emit simulated quotes
            dt = self._freq.astype("timedelta64[ns]").item()
            dt10 = dt // 10

            # - adjust open-close time shift to avoid overlapping timestamps
            if self._open_close_time_shift_secs * S1 >= (dt // 2 - dt10):
                self._open_close_time_shift_secs = (dt // 2 - 2 * dt10) // S1

            if dt < D1:
                self._t_start = self._open_close_time_shift_secs * S1
                self._t_mid1 = dt // 2 - dt10
                self._t_mid2 = dt // 2 + dt10
                self._t_end = dt - self._open_close_time_shift_secs * S1
            else:
                self._t_start = self._d_session_start + self._open_close_time_shift_secs * S1
                self._t_mid1 = dt // 2 - H1
                self._t_mid2 = dt // 2 + H1
                self._t_end = self._d_session_end - self._open_close_time_shift_secs * S1

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer = []
        # - it will fail if receive data doesn't look as ohlcv
        self._time_idx = _find_time_col_idx(column_names)
        self._open_idx = _find_column_index_in_list(column_names, "open")
        self._high_idx = _find_column_index_in_list(column_names, "high")
        self._low_idx = _find_column_index_in_list(column_names, "low")
        self._close_idx = _find_column_index_in_list(column_names, "close")
        self._volume_idx = None
        self._freq = None
        try:
            self._volume_idx = _find_column_index_in_list(column_names, "volume", "vol")
        except:  # noqa: E722
            pass


class RestoreTicksFromOHLC(RestoredEmulatorHelper):
    """
    Emulates quotes (and trades) from OHLC bars
    """

    def __init__(
        self,
        trades: bool = False,  # if we also wants 'trades'
        default_bid_size=1e9,  # default bid/ask is big
        default_ask_size=1e9,  # default bid/ask is big
        daily_session_start_end=DEFAULT_DAILY_SESSION,
        timestamp_units="ns",
        spread=0.0,
        open_close_time_shift_secs=1.0,
        quotes=True,
    ):
        super().__init__(daily_session_start_end, timestamp_units, open_close_time_shift_secs)
        assert trades or quotes or trades and trades, "Either trades or quotes or both must be enabled"
        self._trades = trades
        self._quotes = quotes
        self._bid_size = default_bid_size
        self._ask_size = default_ask_size
        self._s2 = spread / 2.0

    def start_transform(self, name: str, column_names: list[str], **kwargs):
        super().start_transform(name, column_names, **kwargs)
        # -  disable trades if no volume information is available
        if self._volume_idx is None and self._trades:
            logger.warning("Input OHLC data doesn't contain volume information so trades can't be emulated !")
            self._trades = False

    def process_data(self, rows_data: list[list]) -> Any:
        if rows_data is None:
            return

        s2 = self._s2
        if self._freq is None:
            self._detect_emulation_timestamps(rows_data[:100])

        # - input data
        for data in rows_data:
            # ti = pd.Timestamp(data[self._time_idx]).as_unit("ns").asm8.item()
            ti = _time(data[self._time_idx], self._timestamp_units)
            o = data[self._open_idx]
            h = data[self._high_idx]
            l = data[self._low_idx]
            c = data[self._close_idx]
            rv = data[self._volume_idx] if self._volume_idx else 0
            rv = rv / (h - l) if h > l else rv

            # - opening quote
            if self._quotes:
                self.buffer.append(Quote(ti + self._t_start, o - s2, o + s2, self._bid_size, self._ask_size))

            if c >= o:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o - s2, rv * (o - l)))  # sell 1

                if self._quotes:
                    self.buffer.append(
                        Quote(
                            ti + self._t_mid1,
                            l - s2,
                            l + s2,
                            self._bid_size,
                            self._ask_size,
                        )
                    )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, l + s2, rv * (c - o)))  # buy 1

                if self._quotes:
                    self.buffer.append(
                        Quote(
                            ti + self._t_mid2,
                            h - s2,
                            h + s2,
                            self._bid_size,
                            self._ask_size,
                        )
                    )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, h - s2, rv * (h - c)))  # sell 2
            else:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o + s2, rv * (h - o)))  # buy 1

                if self._quotes:
                    self.buffer.append(
                        Quote(
                            ti + self._t_mid1,
                            h - s2,
                            h + s2,
                            self._bid_size,
                            self._ask_size,
                        )
                    )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, h - s2, rv * (o - c)))  # sell 1

                if self._quotes:
                    self.buffer.append(
                        Quote(
                            ti + self._t_mid2,
                            l - s2,
                            l + s2,
                            self._bid_size,
                            self._ask_size,
                        )
                    )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, l + s2, rv * (c - l)))  # buy 2

            # - closing quote
            if self._quotes:
                self.buffer.append(Quote(ti + self._t_end, c - s2, c + s2, self._bid_size, self._ask_size))


class RestoreQuotesFromOHLC(RestoreTicksFromOHLC):
    """
    Restore (emulate) quotes from OHLC bars
    """

    def __init__(
        self,
        default_bid_size=1e9,  # default bid/ask is big
        default_ask_size=1e9,  # default bid/ask is big
        daily_session_start_end=DEFAULT_DAILY_SESSION,
        timestamp_units="ns",
        spread=0.0,
        open_close_time_shift_secs=1.0,
    ):
        super().__init__(
            trades=False,
            default_bid_size=default_bid_size,
            default_ask_size=default_ask_size,
            daily_session_start_end=daily_session_start_end,
            timestamp_units=timestamp_units,
            spread=spread,
            open_close_time_shift_secs=open_close_time_shift_secs,
            quotes=True,
        )


class RestoreTradesFromOHLC(RestoreTicksFromOHLC):
    """
    Restore (emulate) trades from OHLC bars
    """

    def __init__(
        self,
        daily_session_start_end=DEFAULT_DAILY_SESSION,
        timestamp_units="ns",
        open_close_time_shift_secs=1.0,
    ):
        super().__init__(
            trades=True,
            default_bid_size=0,
            default_ask_size=0,
            daily_session_start_end=daily_session_start_end,
            timestamp_units=timestamp_units,
            spread=0,
            open_close_time_shift_secs=open_close_time_shift_secs,
            quotes=False,
        )


class RestoredBarsFromOHLC(RestoredEmulatorHelper):
    """
    Transforms OHLC data into a sequence of bars trying to mimic real-world market data updates
    """

    def __init__(
        self, daily_session_start_end=DEFAULT_DAILY_SESSION, timestamp_units="ns", open_close_time_shift_secs=1.0
    ):
        super().__init__(daily_session_start_end, timestamp_units, open_close_time_shift_secs)

    def process_data(self, rows_data: List[List]) -> Any:
        if rows_data is None:
            return

        if self._freq is None:
            self._detect_emulation_timestamps(rows_data[:100])

        # - input data
        for data in rows_data:
            ti = _time(data[self._time_idx], self._timestamp_units)
            o = data[self._open_idx]
            h = data[self._high_idx]
            l = data[self._low_idx]
            c = data[self._close_idx]

            vol = data[self._volume_idx] if self._volume_idx is not None else 0
            rvol = vol / (h - l) if h > l else vol

            # - opening bar (o,h,l,c=o, v=0)
            self.buffer.append(Bar(ti + self._t_start, o, o, o, o, 0))

            if c >= o:
                v1 = rvol * (o - l)
                self.buffer.append(Bar(ti + self._t_mid1, o, o, l, l, v1))

                v2 = v1 + rvol * (c - o)
                self.buffer.append(Bar(ti + self._t_mid2, o, h, l, h, v2))

            else:
                v1 = rvol * (h - o)
                self.buffer.append(Bar(ti + self._t_mid1, o, h, o, h, v1))

                v2 = v1 + rvol * (o - c)
                self.buffer.append(Bar(ti + self._t_mid2, o, h, l, l, v2))

            # - full bar
            self.buffer.append(Bar(ti + self._t_end, o, h, l, c, vol))


class AsDict(DataTransformer):
    """
    Tries to keep incoming data as list of dictionaries with preprocessed time
    """

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer = list()
        self._time_idx = _find_time_col_idx(column_names)
        self._column_names = column_names
        self._time_name = column_names[self._time_idx]

    def process_data(self, rows_data: Iterable):
        if rows_data is not None:
            for d in rows_data:
                _r_dict = dict(zip(self._column_names, d))
                self.buffer.append(TimestampedDict(_time(d[self._time_idx], "ns"), _r_dict))  # type: ignore


def _retry(fn):
    @wraps(fn)
    def wrapper(*args, **kw):
        cls = args[0]
        for x in range(cls._reconnect_tries):
            # print(x, cls._reconnect_tries)
            try:
                return fn(*args, **kw)
            except (pg.InterfaceError, pg.OperationalError, AttributeError) as e:
                logger.debug("Database Connection [InterfaceError or OperationalError]")
                # print ("Idle for %s seconds" % (cls._reconnect_idle))
                # time.sleep(cls._reconnect_idle)
                cls._connect()

    return wrapper


class QuestDBSqlBuilder:
    """
    Generic sql builder for QuestDB data
    """

    _aliases = {"um": "umfutures", "cm": "cmfutures", "f": "futures"}

    def get_table_name(self, data_id: str, sfx: str = "") -> str:
        """
        Get table name for data_id
        data_id can have format <exchange>.<type>:<symbol>
        for example:
            BINANCE.UM:BTCUSDT or BINANCE:BTCUSDT for spot
        """
        sfx = sfx or "candles_1m"
        table_name = data_id
        _exch, _symb, _mktype = self._get_exchange_symbol_market_type(data_id)
        if _exch and _symb:
            parts = [_exch.lower(), _mktype]
            if "candles" not in sfx:
                parts.append(_symb)
            parts.append(sfx)
            table_name = ".".join(filter(lambda x: x, parts))

        return table_name

    def _get_exchange_symbol_market_type(self, data_id: str) -> tuple[str | None, str | None, str | None]:
        _ss = data_id.split(":")
        if len(_ss) > 1:
            _exch, symb = _ss
            _mktype = "spot"
            _ss = _exch.split(".")
            if len(_ss) > 1:
                _exch = _ss[0]
                _mktype = _ss[1]
            _mktype = _mktype.lower()
            return _exch.lower(), symb.lower(), self._aliases.get(_mktype, _mktype)
        return None, None, None

    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str | None,
        data_type: str,
    ) -> str | None:
        pass

    def prepare_names_sql(self) -> str:
        return "select table_name from tables()"

    def prepare_symbols_sql(self, exchange: str, dtype: str) -> str:
        _table = self.get_table_name(f"{exchange}:BTCUSDT", dtype)
        return f"select distinct(symbol) from {_table}"

    def prepare_data_ranges_sql(self, data_id: str) -> str:
        raise NotImplementedError()


class QuestDBSqlCandlesBuilder(QuestDBSqlBuilder):
    """
    Sql builder for candles data
    """

    def prepare_names_sql(self) -> str:
        return "select table_name from tables() where table_name like '%candles%'"

    @staticmethod
    def _convert_time_delta_to_qdb_resample_format(c_tf: str):
        if c_tf:
            _t = re.match(r"(\d+)(\w+)", c_tf)
            if _t and len(_t.groups()) > 1:
                c_tf = f"{_t[1]}{_t[2][0].lower()}"
        return c_tf

    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str | None,
        data_type: str,
    ) -> str:
        _exch, _symb, _mktype = self._get_exchange_symbol_market_type(data_id)
        if _symb is None:
            _symb = data_id

        where = f"where symbol = '{_symb}'"
        w0 = f"timestamp >= '{start}'" if start else ""
        w1 = f"timestamp < '{end}'" if end else ""

        # - fix: when no data ranges are provided we must skip empy where keyword
        if w0 or w1:
            where = f"{where} and {w0} and {w1}" if (w0 and w1) else f"{where} and {(w0 or w1)}"

        # - filter out candles without any volume
        where = f"{where} and volume > 0"

        # - check resample format
        resample = (
            QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(resample)
            if resample
            else "1m"  # if resample is empty let's use 1 minute timeframe
        )
        _rsmpl = f"SAMPLE by {resample} FILL(NONE)" if resample else ""

        table_name = self.get_table_name(data_id, data_type)
        return f"""
                select timestamp, 
                first(open) as open, max(high) as high, min(low) as low, last(close) as close,
                sum(volume) as volume,
                sum(quote_volume) as quote_volume,
                sum(count) as count,
                sum(taker_buy_volume) as taker_buy_volume,
                sum(taker_buy_quote_volume) as taker_buy_quote_volume
                from "{table_name}" {where} {_rsmpl};
            """

    def prepare_data_ranges_sql(self, data_id: str) -> str:
        _exch, _symb, _mktype = self._get_exchange_symbol_market_type(data_id)
        if _exch is None:
            raise ValueError(f"Can't get exchange name from data id: {data_id} !")
        return f"""(SELECT timestamp FROM "{_exch}.{_mktype}.candles_1m" WHERE symbol='{_symb}' ORDER BY timestamp ASC LIMIT 1)
                        UNION
                   (SELECT timestamp FROM "{_exch}.{_mktype}.candles_1m" WHERE symbol='{_symb}' ORDER BY timestamp DESC LIMIT 1)
                """


class QuestDBSqlTOBBilder(QuestDBSqlBuilder):
    def prepare_data_ranges_sql(self, data_id: str) -> str:
        _exch, _symb, _mktype = self._get_exchange_symbol_market_type(data_id)
        if _exch is None:
            raise ValueError(f"Can't get exchange name from data id: {data_id} !")
        # TODO: ????
        return f"""(SELECT timestamp FROM "{_exch}.{_mktype}.{_symb}.orderbook" ORDER BY timestamp ASC LIMIT 1)
                        UNION
                   (SELECT timestamp FROM "{_exch}.{_mktype}.{_symb}.orderbook" ORDER BY timestamp DESC LIMIT 1)
                """


class QuestDBConnector(DataReader):
    """
    Very first version of QuestDB connector

    ### Connect to an existing QuestDB instance
    >>> db = QuestDBConnector()
    >>> db.read('BINANCE.UM:ETHUSDT', '2024-01-01', transform=AsPandasFrame())
    """

    _reconnect_tries = 5
    _reconnect_idle = 0.1  # wait seconds before retying
    _builder: QuestDBSqlBuilder

    def __init__(
        self,
        builder: QuestDBSqlBuilder = QuestDBSqlCandlesBuilder(),
        host="localhost",
        user="admin",
        password="quest",
        port=8812,
    ) -> None:
        self._connection = None
        self._host = host
        self._port = port
        self.connection_url = f"user={user} password={password} host={host} port={port}"
        self._builder = builder
        self._connect()

    def __getstate__(self):
        if self._connection:
            self._connection.close()
            self._connection = None
        state = self.__dict__.copy()
        return state

    def _connect(self):
        self._connection = pg.connect(self.connection_url, autocommit=True)
        logger.debug(f"Connected to QuestDB at {self._host}:{self._port}")

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timeframe: str | None = "1m",
        data_type="candles_1m",
    ) -> Any:
        return self._read(
            data_id,
            start,
            stop,
            transform,
            chunksize,
            timeframe,
            data_type,
            self._builder,
        )

    def get_candles(
        self,
        exchange: str,
        symbols: list[str],
        start: str | pd.Timestamp,
        stop: str | pd.Timestamp,
        timeframe: str = "1d",
    ) -> pd.DataFrame:
        assert len(symbols) > 0, "No symbols provided"
        quoted_symbols = [f"'{s.lower()}'" for s in symbols]
        where = f"where symbol in ({', '.join(quoted_symbols)}) and timestamp >= '{start}' and timestamp < '{stop}'"
        table_name = QuestDBSqlCandlesBuilder().get_table_name(f"{exchange}:{list(symbols)[0]}")

        _rsmpl = f"sample by {QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(timeframe)}"

        query = f"""
        select timestamp, 
        upper(symbol) as symbol,
        first(open) as open, 
        max(high) as high,
        min(low) as low,
        last(close) as close,
        sum(volume) as volume,
        sum(quote_volume) as quote_volume,
        sum(count) as count,
        sum(taker_buy_volume) as taker_buy_volume,
        sum(taker_buy_quote_volume) as taker_buy_quote_volume
        from "{table_name}" {where} {_rsmpl};
        """
        res = self.execute(query)
        if res.empty:
            return res
        return res.set_index(["timestamp", "symbol"])

    def get_average_quote_volume(
        self,
        exchange: str,
        start: str | pd.Timestamp,
        stop: str | pd.Timestamp,
        timeframe: str = "1d",
    ) -> pd.Series:
        table_name = QuestDBSqlCandlesBuilder().get_table_name(f"{exchange}:BTCUSDT")
        query = f"""
        WITH sampled as (
            select timestamp, symbol, sum(quote_volume) as qvolume 
            from "{table_name}"
            where timestamp >= '{start}' and timestamp < '{stop}'
            SAMPLE BY {QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(timeframe)}
        )
        select upper(symbol) as symbol, avg(qvolume) as quote_volume from sampled
        group by symbol
        order by quote_volume desc;
        """
        vol_stats = self.execute(query)
        if vol_stats.empty:
            return pd.Series()
        return vol_stats.set_index("symbol")["quote_volume"]

    def get_fundamental_data(
        self,
        exchange: str,
        symbols: list[str] | None = None,
        start: str | pd.Timestamp | None = None,
        stop: str | pd.Timestamp | None = None,
        timeframe: str = "1d",
    ) -> pd.DataFrame:
        table_name = {"BINANCE.UM": "binance.umfutures.fundamental"}[exchange]
        query = f"select timestamp, symbol, metric, last(value) as value from {table_name}"
        # TODO: fix handling without start/stop, where needs to be added
        if start or stop:
            conditions = []
            if start:
                conditions.append(f"timestamp >= '{start}'")
            if stop:
                conditions.append(f"timestamp < '{stop}'")
            query += " where " + " and ".join(conditions)
        if symbols:
            query += f" and symbol in ({', '.join(symbols)})"
        _rsmpl = f"sample by {QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(timeframe)}"
        query += f" {_rsmpl}"
        df = self.execute(query)
        if df.empty:
            return pd.DataFrame()
        return df.set_index(["timestamp", "symbol", "metric"]).value.unstack("metric")

    def get_names(self) -> List[str]:
        return self._get_names(self._builder)

    @_retry
    def execute(self, query: str) -> pd.DataFrame:
        _cursor = self._connection.cursor()  # type: ignore
        _cursor.execute(query)  # type: ignore
        names = [d.name for d in _cursor.description]  # type: ignore
        records = _cursor.fetchall()
        if not records:
            return pd.DataFrame()
        return pd.DataFrame(records, columns=names)

    @_retry
    def _read(
        self,
        data_id: str,
        start: str | None,
        stop: str | None,
        transform: DataTransformer,
        chunksize: int,
        timeframe: str | None,
        data_type: str,
        builder: QuestDBSqlBuilder,
    ) -> Any:
        start, end = handle_start_stop(start, stop)
        _req = builder.prepare_data_sql(data_id, start, end, timeframe, data_type)

        _cursor = self._connection.cursor()  # type: ignore
        _cursor.execute(_req)  # type: ignore
        names = [d.name for d in _cursor.description]  # type: ignore

        if chunksize > 0:

            def _iter_chunks():
                while True:
                    records = _cursor.fetchmany(chunksize)
                    if not records:
                        _cursor.close()
                        break
                    transform.start_transform(data_id, names, start=start, stop=stop)
                    transform.process_data(records)
                    yield transform.collect()

            return _iter_chunks()

        try:
            records = _cursor.fetchall()
            if not records:
                return None
            transform.start_transform(data_id, names, start=start, stop=stop)
            transform.process_data(records)
            return transform.collect()
        finally:
            _cursor.close()

    @_retry
    def _get_names(self, builder: QuestDBSqlBuilder) -> list[str]:
        _cursor = None
        try:
            _cursor = self._connection.cursor()  # type: ignore
            _cursor.execute(builder.prepare_names_sql())  # type: ignore
            records = _cursor.fetchall()
        finally:
            if _cursor:
                _cursor.close()
        return [r[0] for r in records]

    @_retry
    def _get_symbols(self, builder: QuestDBSqlBuilder, exchange: str, dtype: str) -> list[str]:
        _cursor = None
        try:
            _cursor = self._connection.cursor()  # type: ignore
            _cursor.execute(builder.prepare_symbols_sql(exchange, dtype))  # type: ignore
            records = _cursor.fetchall()
        finally:
            if _cursor:
                _cursor.close()
        return [f"{exchange}:{r[0].upper()}" for r in records]

    @_retry
    def _get_range(self, builder: QuestDBSqlBuilder, data_id: str) -> tuple[Any] | None:
        _cursor = None
        try:
            _cursor = self._connection.cursor()  # type: ignore
            _cursor.execute(builder.prepare_data_ranges_sql(data_id))  # type: ignore
            return tuple([np.datetime64(r[0]) for r in _cursor.fetchall()])
        finally:
            if _cursor:
                _cursor.close()

    def __del__(self):
        try:
            if self._connection is not None:
                logger.debug("Closing connection")
                self._connection.close()
        except:  # noqa: E722
            pass


class QuestDBSqlOrderBookBuilder(QuestDBSqlCandlesBuilder):
    """
    Sql builder for snapshot data
    """

    SNAPSHOT_DELTA = pd.Timedelta("1h")
    MIN_DELTA = pd.Timedelta("1s")

    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str,
        data_type: str,
    ) -> str:
        if not start or not end:
            raise ValueError("Start and end dates must be provided for orderbook data!")
        start_dt, end_dt = pd.Timestamp(start), pd.Timestamp(end)
        delta = end_dt - start_dt

        raw_start_dt = start_dt.floor(self.SNAPSHOT_DELTA) - self.MIN_DELTA

        table_name = self.get_table_name(data_id, data_type)
        query = f"""
SELECT * FROM {table_name}
WHERE timestamp BETWEEN '{raw_start_dt}' AND '{end_dt}'
"""
        return query


class TradeSql(QuestDBSqlCandlesBuilder):
    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str,
        data_type: str,
    ) -> str:
        table_name = self.get_table_name(data_id, data_type)
        where = ""
        w0 = f"timestamp >= '{start}'" if start else ""
        w1 = f"timestamp <= '{end}'" if end else ""

        # - fix: when no data ranges are provided we must skip empy where keyword
        if w0 or w1:
            where = f"where {w0} and {w1}" if (w0 and w1) else f"where {(w0 or w1)}"

        resample = (
            QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(resample) if resample else resample
        )
        if resample:
            sql = f"""
                select timestamp, first(price) as open, max(price) as high, min(price) as low, last(price) as close, 
                sum(size) as volume from "{table_name}" {where} SAMPLE by {resample};"""
        else:
            sql = f"""select timestamp, price, size, market_maker from "{table_name}" {where};"""

        return sql

    def prepare_symbols_sql(self, exchange: str, dtype: str) -> str:
        # TODO:
        raise NotImplementedError("Not implemented yet")


class MultiQdbConnector(QuestDBConnector):
    """
    Data connector for QuestDB which provides access to following data types:
      - candles
      - trades
      - orderbook snapshots
      - liquidations
      - funding rate

    Examples:
    1. Retrieving trades:
        qdb.read(
            "BINANCE.UM:BTCUSDT",
            "2023-01-01 00:00",
            "2023-01-01 10:00",
            timeframe="15Min",
            transform=AsPandasFrame(),
            data_type="trade"
        )
    """

    _TYPE_TO_BUILDER = {
        "candles_1m": QuestDBSqlCandlesBuilder(),
        "tob": QuestDBSqlTOBBilder(),
        "trade": TradeSql(),
        "agg_trade": TradeSql(),
        "orderbook": QuestDBSqlOrderBookBuilder(),
    }

    _TYPE_MAPPINGS = {
        "candles": "candles_1m",
        "ohlc": "candles_1m",
        "trades": "trade",
        "ob": "orderbook",
        "trd": "trade",
        "td": "trade",
        "quote": "tob",
        "aggTrade": "agg_trade",
        "agg_trades": "agg_trade",
        "aggTrades": "agg_trade",
    }

    def __init__(
        self,
        host="localhost",
        user="admin",
        password="quest",
        port=8812,
    ) -> None:
        self._connection = None
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._connect()

    @property
    def connection_url(self):
        return " ".join(
            [
                f"user={self._user}",
                f"password={self._password}",
                f"host={self._host}",
                f"port={self._port}",
            ]
        )

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize: int = 0,
        timeframe: str | None = None,
        data_type: str = "candles",
    ) -> Any:
        _mapped_data_type = self._TYPE_MAPPINGS.get(data_type, data_type)
        return self._read(
            data_id,
            start,
            stop,
            transform,
            chunksize,
            timeframe,
            _mapped_data_type,
            self._TYPE_TO_BUILDER[_mapped_data_type],
        )

    def get_names(self, data_type: str) -> list[str]:
        return self._get_names(self._TYPE_TO_BUILDER[self._TYPE_MAPPINGS.get(data_type, data_type)])

    def get_symbols(self, exchange: str, dtype: str) -> list[str]:
        return self._get_symbols(
            self._TYPE_TO_BUILDER[self._TYPE_MAPPINGS.get(dtype, dtype)],
            exchange,
            self._TYPE_MAPPINGS.get(dtype, dtype),
        )

    def get_time_ranges(self, symbol: str, dtype: str) -> tuple[np.datetime64, np.datetime64]:
        try:
            _xr = self._get_range(self._TYPE_TO_BUILDER[self._TYPE_MAPPINGS.get(dtype, dtype)], symbol)
            return (None, None) if not _xr else _xr  # type: ignore
        except Exception:
            return (None, None)  # type: ignore



import traceback
from typing import Any

import ccxt
import ccxt.pro as cxp
from ccxt.base.errors import ExchangeError
from qubx import logger
from qubx.core.basics import (
    CtrlChannel,
    Instrument,
    Order,
    Position,
)
from qubx.core.interfaces import (
    IAccountProcessor,
    IBroker,
    ITimeProvider,
)
from qubx.utils.misc import AsyncThreadLoop

from .utils import ccxt_convert_order_info, instrument_to_ccxt_symbol


class CcxtBroker(IBroker):
    _exchange: cxp.Exchange

    _positions: dict[Instrument, Position]
    _loop: AsyncThreadLoop

    def __init__(
        self,
        exchange: cxp.Exchange,
        channel: CtrlChannel,
        time_provider: ITimeProvider,
        account: IAccountProcessor,
    ):
        self._exchange = exchange
        self.ccxt_exchange_id = str(exchange.name)
        self.channel = channel
        self.time_provider = time_provider
        self.account = account
        self._loop = AsyncThreadLoop(exchange.asyncio_loop)

    @property
    def is_simulated_trading(self) -> bool:
        return False

    def send_order(
        self,
        instrument: Instrument,
        order_side: str,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
    ) -> Order:
        params = {}

        if order_type == "limit":
            params["timeInForce"] = time_in_force.upper()
            if price is None:
                raise ValueError("Price must be specified for limit order")

        if client_id:
            params["newClientOrderId"] = client_id

        if instrument.is_futures():
            params["type"] = "swap"

        ccxt_symbol = instrument_to_ccxt_symbol(instrument)

        r: dict[str, Any] | None = None
        try:
            r = self._loop.submit(
                self._exchange.create_order(
                    symbol=ccxt_symbol,
                    type=order_type,  # type: ignore
                    side=order_side,  # type: ignore
                    amount=amount,
                    price=price,
                    params=params,
                )
            ).result()
        except ccxt.BadRequest as exc:
            logger.error(
                f"(::send_order) BAD REQUEST for {order_side} {amount} {order_type} for {instrument.symbol} : {exc}"
            )
            raise exc
        except Exception as err:
            logger.error(f"(::send_order) {order_side} {amount} {order_type} for {instrument.symbol} exception : {err}")
            logger.error(traceback.format_exc())
            raise err

        if r is None:
            msg = "(::send_order) No response from exchange"
            logger.error(msg)
            raise ExchangeError(msg)

        order = ccxt_convert_order_info(instrument, r)
        logger.info(f"New order {order}")
        return order

    def cancel_order(self, order_id: str) -> Order | None:
        order = None
        orders = self.account.get_orders()
        if order_id in orders:
            order = orders[order_id]
            try:
                logger.info(f"Canceling order {order_id} ...")
                result = self._loop.submit(
                    self._exchange.cancel_order(order_id, symbol=instrument_to_ccxt_symbol(order.instrument))
                ).result()
                logger.debug(f"Cancel order result: {result}")
                return order
            except Exception as err:
                logger.error(f"Canceling [{order}] exception : {err}")
                logger.error(traceback.format_exc())
                raise err
        return order

    def cancel_orders(self, instrument: Instrument) -> None:
        raise NotImplementedError("Not implemented yet")

    def update_order(self, order_id: str, price: float | None = None, amount: float | None = None) -> Order:
        raise NotImplementedError("Not implemented yet")

    def exchange(self) -> str:
        """
        Return the name of the exchange this broker is connected to.
        """
        return self.ccxt_exchange_id.upper()



import os
import shutil
import subprocess
import zipfile
from pathlib import Path

from qubx import logger


def validate_zip_file(zip_file: str) -> bool:
    """
    Validates that the provided file is a zip file.

    Args:
        zip_file: Path to the zip file to validate

    Returns:
        bool: True if valid, False otherwise
    """
    if not zip_file.endswith(".zip"):
        logger.error("The file must be a zip file with .zip extension")
        return False
    return True


def determine_output_directory(zip_file: str, output_dir: str | None) -> str:
    """
    Determines the output directory for the deployment.

    Args:
        zip_file: Path to the zip file to deploy
        output_dir: User-specified output directory or None

    Returns:
        str: The resolved output directory path
    """
    zip_path = Path(zip_file)
    zip_name = zip_path.stem

    if output_dir is None:
        return str(zip_path.parent / zip_name)

    output_dir = os.path.abspath(os.path.expanduser(output_dir))
    # If output_dir is a directory that exists, create a subdirectory with the zip name
    if os.path.isdir(output_dir):
        return os.path.join(output_dir, zip_name)

    return output_dir


def prepare_output_directory(output_dir: str, force: bool) -> bool:
    """
    Prepares the output directory, handling existing directories based on the force flag.

    Args:
        output_dir: The output directory path
        force: Whether to force overwrite if the directory exists

    Returns:
        bool: True if successful, False otherwise
    """
    if os.path.exists(output_dir):
        if not force:
            logger.error(f"Output directory {output_dir} already exists. Use --force to overwrite.")
            return False
        logger.warning(f"Removing existing directory {output_dir}")
        shutil.rmtree(output_dir)

    os.makedirs(output_dir, exist_ok=True)
    return True


def extract_zip_file(zip_file: str, output_dir: str) -> bool:
    """
    Extracts the zip file to the output directory.

    Args:
        zip_file: Path to the zip file to extract
        output_dir: Directory to extract to

    Returns:
        bool: True if successful, False otherwise
    """
    logger.info(f"Unpacking {zip_file} to {output_dir}")
    try:
        with zipfile.ZipFile(zip_file, "r") as zip_ref:
            zip_ref.extractall(output_dir)
        return True
    except zipfile.BadZipFile:
        logger.error(f"The file {zip_file} is not a valid zip file")
        return False
    except Exception as e:
        logger.error(f"Failed to unpack zip file: {e}")
        return False


def ensure_poetry_lock_exists(output_dir: str) -> bool:
    """
    Ensures that a poetry.lock file exists in the output directory.
    If not, attempts to generate one.

    Args:
        output_dir: The directory to check/generate in

    Returns:
        bool: True if successful, False otherwise
    """
    poetry_lock_path = os.path.join(output_dir, "poetry.lock")
    if not os.path.exists(poetry_lock_path):
        logger.warning("poetry.lock not found in the zip file. Attempting to generate it.")
        try:
            subprocess.run(
                ["poetry", "lock", "--no-update"], cwd=output_dir, check=True, capture_output=True, text=True
            )
            return True
        except subprocess.CalledProcessError as e:
            logger.error(f"Failed to generate poetry.lock: {e.stderr}")
            return False
    return True


def setup_poetry_environment(output_dir: str) -> bool:
    """
    Sets up the Poetry virtual environment in the output directory.

    Args:
        output_dir: The directory to set up the environment in

    Returns:
        bool: True if successful, False otherwise
    """
    logger.info("Creating Poetry virtual environment")
    try:
        # Configure Poetry to create a virtual environment in the .venv directory
        logger.info("Configuring Poetry")
        subprocess.run(
            ["poetry", "config", "virtualenvs.in-project", "true", "--local"],
            cwd=output_dir,
            check=True,
            capture_output=True,
            text=True,
        )

        # Check if we're already in a Poetry shell
        in_poetry_env = "POETRY_ACTIVE" in os.environ or "VIRTUAL_ENV" in os.environ

        if in_poetry_env:
            logger.debug(
                "Detected active Poetry environment. "
                "Will explicitly create a new environment for the deployed strategy."
            )

        # Install dependencies
        logger.info("Installing dependencies")

        # If we're in a Poetry shell, we need to be more explicit about creating a new environment
        install_cmd = ["poetry", "install"]
        if in_poetry_env:
            # Force Poetry to create a new environment even if we're in an active one
            env = os.environ.copy()
            # Temporarily unset Poetry environment variables to avoid interference
            for var in ["POETRY_ACTIVE", "VIRTUAL_ENV"]:
                if var in env:
                    del env[var]

            subprocess.run(install_cmd, cwd=output_dir, check=True, capture_output=True, text=True, env=env)
        else:
            # Normal case - not in a Poetry shell
            subprocess.run(install_cmd, cwd=output_dir, check=True, capture_output=True, text=True)

        # Verify that the virtual environment was created
        venv_path = os.path.join(output_dir, ".venv")
        if not os.path.exists(venv_path):
            logger.warning(
                "Virtual environment directory (.venv) not found. "
                "This might happen if you're already in a Poetry shell. "
                "You may need to run 'cd %s && poetry env use python' to create it manually.",
                output_dir,
            )

        return True
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to set up Poetry environment: {e.stderr}")
        return False


def deploy_strategy(zip_file: str, output_dir: str | None, force: bool) -> bool:
    """
    Deploys a strategy from a zip file created by the release command.

    This function:
    1. Unpacks the zip file to the specified output directory
    2. Creates a Poetry virtual environment in the .venv folder
    3. Installs dependencies from the poetry.lock file

    Args:
        zip_file: Path to the zip file to deploy
        output_dir: Output directory to unpack the zip file. If None, uses the directory containing the zip file.
        force: Whether to force overwrite if the output directory already exists

    Returns:
        bool: True if deployment was successful, False otherwise
    """
    # Validate the zip file
    if not validate_zip_file(zip_file):
        return False

    # Determine the output directory
    resolved_output_dir = determine_output_directory(zip_file, output_dir)

    # Prepare the output directory
    if not prepare_output_directory(resolved_output_dir, force):
        return False

    # Extract the zip file
    if not extract_zip_file(zip_file, resolved_output_dir):
        return False

    # Ensure poetry.lock exists
    if not ensure_poetry_lock_exists(resolved_output_dir):
        return False

    # Set up the Poetry environment
    if not setup_poetry_environment(resolved_output_dir):
        return False

    # Success messages
    logger.info(f"Strategy deployed successfully to {resolved_output_dir}")
    logger.info(
        f"To run the strategy (paper mode): <cyan>cd {resolved_output_dir} && poetry run qubx run config.yml --paper</cyan>"
    )
    return True



from pathlib import Path

import yaml
from pydantic import BaseModel, Field


class ExchangeConfig(BaseModel):
    connector: str
    universe: list[str]


class AuxConfig(BaseModel):
    reader: str
    args: dict = Field(default_factory=dict)


class LoggingConfig(BaseModel):
    logger: str
    position_interval: str
    portfolio_interval: str
    heartbeat_interval: str = "1m"


class StrategyConfig(BaseModel):
    strategy: str
    parameters: dict = Field(default_factory=dict)
    exchanges: dict[str, ExchangeConfig]
    logging: LoggingConfig
    aux: AuxConfig | None = None


def load_strategy_config_from_yaml(path: Path | str, key: str | None = None) -> StrategyConfig:
    """
    Loads a strategy configuration from a YAML file.

    Args:
        path (str | Path): The path to the YAML file.
        key (str | None): The key to extract from the YAML file.

    Returns:
        StrategyConfig: The parsed configuration.
    """
    path = Path(path)
    if not path.exists():
        raise FileNotFoundError(f"File {path} not found.")
    with path.open("r") as f:
        config_dict = yaml.safe_load(f)
        if key:
            config_dict = config_dict[key]
        return StrategyConfig(**config_dict)


class StrategySimulationConfig(BaseModel):
    strategy: str | list[str]
    parameters: dict = Field(default_factory=dict)
    data: dict = Field(default_factory=dict)
    simulation: dict = Field(default_factory=dict)
    description: str | list[str] | None = None
    variate: dict = Field(default_factory=dict)


def load_simulation_config_from_yaml(path: Path | str) -> StrategySimulationConfig:
    with open(path, "r") as f:
        cfg = yaml.safe_load(f)
    return StrategySimulationConfig(**cfg)



from typing import Literal

import numpy as np
import pandas as pd
from joblib import delayed

from qubx import QubxLogConfig, logger
from qubx.core.basics import SW, DataType
from qubx.core.context import StrategyContext
from qubx.core.exceptions import SimulationConfigError, SimulationError
from qubx.core.helpers import extract_parameters_from_object, full_qualified_class_name
from qubx.core.interfaces import IStrategy
from qubx.core.loggers import InMemoryLogsWriter, StrategyLogging
from qubx.core.lookups import lookup
from qubx.core.metrics import TradingSessionResult
from qubx.data.readers import DataReader
from qubx.pandaz.utils import _frame_to_str
from qubx.utils.misc import ProgressParallel, Stopwatch, get_current_user
from qubx.utils.time import handle_start_stop

from .account import SimulatedAccountProcessor
from .broker import SimulatedBroker
from .data import SimulatedDataProvider
from .utils import (
    DataDecls_t,
    ExchangeName_t,
    SetupTypes,
    SignalsProxy,
    SimulatedCtrlChannel,
    SimulatedLogFormatter,
    SimulatedScheduler,
    SimulatedTimeProvider,
    SimulationDataConfig,
    SimulationSetup,
    StrategiesDecls_t,
    SymbolOrInstrument_t,
    find_instruments_and_exchanges,
    recognize_simulation_configuration,
    recognize_simulation_data_config,
)


def simulate(
    strategies: StrategiesDecls_t,
    data: DataDecls_t,
    capital: float,
    instruments: list[SymbolOrInstrument_t] | dict[ExchangeName_t, list[SymbolOrInstrument_t]],
    commissions: str,
    start: str | pd.Timestamp,
    stop: str | pd.Timestamp | None = None,
    exchange: ExchangeName_t | None = None,
    base_currency: str = "USDT",
    n_jobs: int = 1,
    silent: bool = False,
    aux_data: DataReader | None = None,
    accurate_stop_orders_execution: bool = False,
    signal_timeframe: str = "1Min",
    open_close_time_indent_secs=1,
    debug: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] | None = "WARNING",
    show_latency_report: bool = False,
    parallel_backend: Literal["loky", "multiprocessing"] = "multiprocessing",
) -> list[TradingSessionResult]:
    """
    Backtest utility for trading strategies or signals using historical data.

    Args:
        - strategies (StrategiesDecls_t): Trading strategy or signals configuration.
        - data (DataDecls_t): Historical data for simulation, either as a dictionary of DataFrames or a DataReader object.
        - capital (float): Initial capital for the simulation.
        - instruments (list[SymbolOrInstrument_t] | dict[ExchangeName_t, list[SymbolOrInstrument_t]]): List of trading instruments or a dictionary mapping exchanges to instrument lists.
        - commissions (str): Commission structure for trades.
        - start (str | pd.Timestamp): Start time of the simulation.
        - stop (str | pd.Timestamp | None): End time of the simulation. If None, simulates until the last accessible data.
        - exchange (ExchangeName_t | None): Exchange name if not specified in the instruments list.
        - base_currency (str): Base currency for the simulation, default is "USDT".
        - n_jobs (int): Number of parallel jobs for simulation, default is 1.
        - silent (bool): If True, suppresses output during simulation.
        - aux_data (DataReader | None): Auxiliary data provider (default is None).
        - accurate_stop_orders_execution (bool): If True, enables more accurate stop order execution simulation.
        - signal_timeframe (str): Timeframe for signals, default is "1Min".
        - open_close_time_indent_secs (int): Time indent in seconds for open/close times, default is 1.
        - debug (Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] | None): Logging level for debugging.
        - show_latency_report: If True, shows simulator's latency report.

    Returns:
        - list[TradingSessionResult]: A list of TradingSessionResult objects containing the results of each simulation setup.
    """

    # - setup logging
    QubxLogConfig.set_log_level(debug.upper() if debug else "WARNING")

    # - we need to reset stopwatch
    Stopwatch().reset()

    # - process instruments:
    _instruments, _exchanges = find_instruments_and_exchanges(instruments, exchange)

    # - check we have exchange
    if not _exchanges:
        logger.error(
            _msg
            := "No exchange information provided - you can specify it by exchange parameter or use <yellow>EXCHANGE:SYMBOL</yellow> format for symbols"
        )
        raise SimulationError(_msg)

    # - check if instruments are from the same exchange (mmulti-exchanges is not supported yet)
    if len(_exchanges) > 1:
        logger.error(
            _msg := f"Multiple exchanges found: {', '.join(_exchanges)} - this mode is not supported yet in Qubx !"
        )
        raise SimulationError(_msg)

    exchange = _exchanges[0]

    # - recognize provided data
    data_setup = recognize_simulation_data_config(data, _instruments, exchange, open_close_time_indent_secs, aux_data)

    # - recognize setup: it can be either a strategy or set of signals
    simulation_setups = recognize_simulation_configuration(
        "",
        strategies,
        _instruments,
        exchange,
        capital,
        base_currency,
        commissions,
        signal_timeframe,
        accurate_stop_orders_execution,
    )
    if not simulation_setups:
        logger.error(
            _msg
            := "Can't recognize setup - it should be a strategy, a set of signals or list of signals/strategies + tracker !"
        )
        raise SimulationError(_msg)

    # - preprocess start and stop and convert to datetime if necessary
    if stop is None:
        # - check stop time : here we try to backtest till now (may be we need to get max available time from data reader ?)
        stop = pd.Timestamp.now(tz="UTC").astimezone(None)

    _start, _stop = handle_start_stop(start, stop, convert=pd.Timestamp)
    assert isinstance(_start, pd.Timestamp) and isinstance(_stop, pd.Timestamp), "Invalid start and stop times"

    # - run simulations
    return _run_setups(
        simulation_setups,
        data_setup,
        _start,
        _stop,
        n_jobs=n_jobs,
        silent=silent,
        show_latency_report=show_latency_report,
        parallel_backend=parallel_backend,
    )


def _run_setups(
    strategies_setups: list[SimulationSetup],
    data_setup: SimulationDataConfig,
    start: pd.Timestamp,
    stop: pd.Timestamp,
    n_jobs: int = -1,
    silent: bool = False,
    show_latency_report: bool = False,
    parallel_backend: Literal["loky", "multiprocessing"] = "multiprocessing",
) -> list[TradingSessionResult]:
    # loggers don't work well with joblib and multiprocessing in general because they contain
    # open file handlers that cannot be pickled. I found a solution which requires the usage of enqueue=True
    # in the logger configuration and specifying backtest "multiprocessing" instead of the default "loky"
    # for joblib. But it works now.
    # See: https://stackoverflow.com/questions/59433146/multiprocessing-logging-how-to-use-loguru-with-joblib-parallel
    _main_loop_silent = len(strategies_setups) == 1
    n_jobs = 1 if _main_loop_silent else n_jobs

    reports = ProgressParallel(
        n_jobs=n_jobs, total=len(strategies_setups), silent=_main_loop_silent, backend=parallel_backend
    )(
        delayed(_run_setup)(id, f"Simulated-{id}", setup, data_setup, start, stop, silent, show_latency_report)
        for id, setup in enumerate(strategies_setups)
    )
    return reports  # type: ignore


def _run_setup(
    setup_id: int,
    account_id: str,
    setup: SimulationSetup,
    data_setup: SimulationDataConfig,
    start: pd.Timestamp,
    stop: pd.Timestamp,
    silent: bool,
    show_latency_report: bool,
) -> TradingSessionResult:
    _stop = pd.Timestamp(stop)

    # - fees for this exchange
    tcc = lookup.fees.find(setup.exchange.lower(), setup.commissions)
    if tcc is None:
        raise SimulationConfigError(
            f"Can't find transaction costs calculator for '{setup.exchange}' for specification '{setup.commissions}' !"
        )

    channel = SimulatedCtrlChannel("databus", sentinel=(None, None, None, None))
    simulated_clock = SimulatedTimeProvider(np.datetime64(start, "ns"))

    # - we want to see simulate time in log messages
    QubxLogConfig.setup_logger(QubxLogConfig.get_log_level(), SimulatedLogFormatter(simulated_clock).formatter)

    logger.debug(
        f"[<y>simulator</y>] :: Preparing simulated trading on <g>{setup.exchange.upper()}</g> for {setup.capital} {setup.base_currency}..."
    )

    account = SimulatedAccountProcessor(
        account_id=account_id,
        channel=channel,
        base_currency=setup.base_currency,
        initial_capital=setup.capital,
        time_provider=simulated_clock,
        tcc=tcc,
        accurate_stop_orders_execution=setup.accurate_stop_orders_execution,
    )
    scheduler = SimulatedScheduler(channel, lambda: simulated_clock.time().item())
    broker = SimulatedBroker(channel, account, setup.exchange)
    data_provider = SimulatedDataProvider(
        exchange_id=setup.exchange,
        channel=channel,
        scheduler=scheduler,
        time_provider=simulated_clock,
        account=account,
        readers=data_setup.data_providers,
        open_close_time_indent_secs=data_setup.adjusted_open_close_time_indent_secs,
    )

    # - it will store simulation results into memory
    logs_writer = InMemoryLogsWriter(account_id, setup.name, "0")
    strat: IStrategy | None = None

    match setup.setup_type:
        case SetupTypes.STRATEGY:
            strat = setup.generator  # type: ignore

        case SetupTypes.STRATEGY_AND_TRACKER:
            strat = setup.generator  # type: ignore
            strat.tracker = lambda ctx: setup.tracker  # type: ignore

        case SetupTypes.SIGNAL:
            strat = SignalsProxy(timeframe=setup.signal_timeframe)
            data_provider.set_generated_signals(setup.generator)  # type: ignore

            # - we don't need any unexpected triggerings
            _stop = min(setup.generator.index[-1], _stop)  # type: ignore

        case SetupTypes.SIGNAL_AND_TRACKER:
            strat = SignalsProxy(timeframe=setup.signal_timeframe)
            strat.tracker = lambda ctx: setup.tracker
            data_provider.set_generated_signals(setup.generator)  # type: ignore

            # - we don't need any unexpected triggerings
            _stop = min(setup.generator.index[-1], _stop)  # type: ignore

        case _:
            raise SimulationError(f"Unsupported setup type: {setup.setup_type} !")

    if not isinstance(strat, IStrategy):
        raise SimulationConfigError(f"Strategy should be an instance of IStrategy, but got {strat} !")

    # - get aux data provider
    _aux_data = data_setup.get_timeguarded_aux_reader(simulated_clock)

    ctx = StrategyContext(
        strategy=strat,
        broker=broker,
        data_provider=data_provider,
        account=account,
        scheduler=scheduler,
        time_provider=simulated_clock,
        instruments=setup.instruments,
        logging=StrategyLogging(logs_writer),
        aux_data_provider=_aux_data,
    )

    # - setup base subscription from spec
    if ctx.get_base_subscription() == DataType.NONE:
        logger.debug(
            f"[<y>simulator</y>] :: Setting up default base subscription: {data_setup.default_base_subscription}"
        )
        ctx.set_base_subscription(data_setup.default_base_subscription)

    # - set default on_event schedule if detected and strategy didn't set it's own schedule
    if not ctx.get_event_schedule("time") and data_setup.default_trigger_schedule:
        logger.debug(f"[<y>simulator</y>] :: Setting default schedule: {data_setup.default_trigger_schedule}")
        ctx.set_event_schedule(data_setup.default_trigger_schedule)

    # - get strategy parameters BEFORE simulation start
    #   potentially strategy may change it's parameters during simulation
    _s_class, _s_params = "", None
    if setup.setup_type in [SetupTypes.STRATEGY, SetupTypes.STRATEGY_AND_TRACKER]:
        _s_params = extract_parameters_from_object(setup.generator)
        _s_class = full_qualified_class_name(setup.generator)

    # - start context at this point
    ctx.start()

    # - apply default warmup periods if strategy didn't set them
    for s in ctx.get_subscriptions():
        if not ctx.get_warmup(s) and (_d_wt := data_setup.default_warmups.get(s)):
            logger.debug(
                f"[<y>simulator</y>] :: Strategy didn't set warmup period for <c>{s}</c> so default <c>{_d_wt}</c> will be used"
            )
            ctx.set_warmup({s: _d_wt})

    def _is_known_type(t: str):
        try:
            DataType(t)
            return True
        except:  # noqa: E722
            return False

    # - if any custom data providers are in the data spec
    for t, r in data_setup.data_providers.items():
        if not _is_known_type(t) or t in [DataType.TRADE, DataType.OHLC_TRADES, DataType.OHLC_QUOTES, DataType.QUOTE]:
            logger.debug(f"[<y>simulator</y>] :: Subscribing to: {t}")
            ctx.subscribe(t, ctx.instruments)

    try:
        data_provider.run(start, _stop, silent=silent)  # type: ignore
    except KeyboardInterrupt:
        logger.error("Simulated trading interrupted by user !")

    # - stop context at this point
    ctx.stop()

    # - service latency report
    if show_latency_report:
        _l_r = SW.latency_report()
        if _l_r is not None:
            logger.info(
                "<BLUE>   Time spent in simulation report   </BLUE>\n<r>"
                + _frame_to_str(
                    _l_r.sort_values("latency", ascending=False).reset_index(drop=True), "simulation", -1, -1, False
                )
                + "</r>"
            )

    return TradingSessionResult(
        setup_id,
        setup.name,
        start,
        stop,
        setup.exchange,
        setup.instruments,
        setup.capital,
        setup.base_currency,
        setup.commissions,
        logs_writer.get_portfolio(as_plain_dataframe=True),
        logs_writer.get_executions(),
        logs_writer.get_signals(),
        strategy_class=_s_class,
        parameters=_s_params,
        is_simulation=True,
        author=get_current_user(),
    )



import os
import sys
from pathlib import Path

import click

from qubx import QubxLogConfig, logger


@click.group()
@click.option(
    "--debug",
    "-d",
    is_flag=True,
    help="Enable debug mode.",
)
@click.option(
    "--debug-port",
    "-p",
    type=int,
    help="Debug port.",
    default=5678,
)
@click.option(
    "--log-level",
    "-l",
    type=str,
    help="Log level.",
    default="INFO",
)
def main(debug: bool, debug_port: int, log_level: str):
    """
    Qubx CLI.
    """
    log_level = log_level.upper() if not debug else "DEBUG"

    QubxLogConfig.set_log_level(log_level)

    if debug:
        os.environ["PYDEVD_DISABLE_FILE_VALIDATION"] = "1"

        import debugpy

        logger.info(f"Waiting for debugger to attach (port {debug_port})")

        debugpy.listen(debug_port)
        debugpy.wait_for_client()


@main.command()
@click.argument("config-file", type=Path, required=True)
@click.option(
    "--account-file",
    "-a",
    type=Path,
    help="Account configuration file path.",
    required=False,
)
@click.option("--paper", "-p", is_flag=True, default=False, help="Use paper trading mode.", show_default=True)
@click.option(
    "--jupyter", "-j", is_flag=True, default=False, help="Run strategy in jupyter console.", show_default=True
)
def run(config_file: Path, account_file: Path | None, paper: bool, jupyter: bool):
    """
    Starts the strategy with the given configuration file. If paper mode is enabled, account is not required.

    Account configurations are searched in the following priority:\n
    - If provided, the account file is searched first.\n
    - If exists, accounts.toml located in the same folder with the config searched.\n
    - If neither of the above are provided, the accounts.toml in the ~/qubx/accounts.toml path is searched.
    """
    from qubx.utils.misc import add_project_to_system_path, logo
    from qubx.utils.runner.runner import run_strategy_yaml, run_strategy_yaml_in_jupyter

    add_project_to_system_path()
    add_project_to_system_path(str(config_file.parent))
    if jupyter:
        run_strategy_yaml_in_jupyter(config_file, account_file, paper)
    else:
        logo()
        run_strategy_yaml(config_file, account_file, paper, blocking=True)


@main.command()
@click.argument("config-file", type=Path, required=True)
@click.option(
    "--start", "-s", default=None, type=str, help="Override simulation start date from config.", show_default=True
)
@click.option(
    "--end", "-e", default=None, type=str, help="Override simulation end date from config.", show_default=True
)
@click.option(
    "--output", "-o", default="results", type=str, help="Output directory for simulation results.", show_default=True
)
def simulate(config_file: Path, start: str | None, end: str | None, output: str | None):
    """
    Simulates the strategy with the given configuration file.
    """
    from qubx.utils.misc import add_project_to_system_path, logo
    from qubx.utils.runner.runner import simulate_strategy

    add_project_to_system_path()
    add_project_to_system_path(str(config_file.parent))
    logo()
    simulate_strategy(config_file, output, start, end)


@main.command()
@click.argument(
    "directory",
    type=click.Path(exists=True, resolve_path=True),
    default=".",
    callback=lambda ctx, param, value: os.path.abspath(os.path.expanduser(value)),
)
def ls(directory: str):
    """
    Lists all strategies in the given directory.

    Strategies are identified by the inheritance from IStrategy interface.
    """
    from .release import ls_strats

    ls_strats(directory)


@main.command()
@click.argument(
    "directory",
    type=click.Path(exists=False),
    default=".",
    callback=lambda ctx, param, value: os.path.abspath(os.path.expanduser(value)),
)
@click.option(
    "--strategy",
    "-s",
    type=click.STRING,
    help="Strategy name to release (should match the strategy class name) or path to a config YAML file",
    required=True,
)
@click.option(
    "--output-dir",
    "-o",
    type=click.STRING,
    help="Output directory to put zip file.",
    default="releases",
    show_default=True,
)
@click.option(
    "--tag",
    "-t",
    type=click.STRING,
    help="Additional tag for this release (e.g. 'v1.0.0')",
    required=False,
)
@click.option(
    "--message",
    "-m",
    type=click.STRING,
    help="Release message (added to the info yaml file).",
    required=False,
    default=None,
    show_default=True,
)
@click.option(
    "--commit",
    "-c",
    is_flag=True,
    default=False,
    help="Commit changes and create tag in repo (default: False)",
    show_default=True,
)
@click.option(
    "--default-exchange",
    type=click.STRING,
    help="Default exchange to use in the generated config.",
    default="BINANCE.UM",
    show_default=True,
)
@click.option(
    "--default-connector",
    type=click.STRING,
    help="Default connector to use in the generated config.",
    default="ccxt",
    show_default=True,
)
@click.option(
    "--default-instruments",
    type=click.STRING,
    help="Default instruments to use in the generated config (comma-separated).",
    default="BTCUSDT",
    show_default=True,
)
def release(
    directory: str,
    strategy: str,
    tag: str | None,
    message: str | None,
    commit: bool,
    output_dir: str,
    default_exchange: str,
    default_connector: str,
    default_instruments: str,
) -> None:
    """
    Releases the strategy to a zip file.

    The strategy can be specified in two ways:
    1. As a strategy name (class name) - strategies are scanned in the given directory
    2. As a path to a config YAML file containing the strategy configuration in StrategyConfig format

    If a strategy name is provided, a default configuration will be generated with:
    - The strategy parameters from the strategy class
    - Default exchange, connector, and instruments from the command options
    - Standard logging configuration

    If a config file is provided, it must follow the StrategyConfig structure with:
    - strategy: The strategy name or path
    - parameters: Dictionary of strategy parameters
    - exchanges: Dictionary of exchange configurations
    - aux: Auxiliary configuration
    - logging: Logging configuration

    All of the dependencies are included in the zip file.
    """
    from .release import release_strategy

    # Parse default instruments
    instruments = [instr.strip() for instr in default_instruments.split(",")]

    release_strategy(
        directory=directory,
        strategy_name=strategy,
        tag=tag,
        message=message,
        commit=commit,
        output_dir=output_dir,
        default_exchange=default_exchange,
        default_connector=default_connector,
        default_instruments=instruments,
    )


@main.command()
@click.argument(
    "zip-file",
    type=click.Path(exists=True, resolve_path=True),
    callback=lambda ctx, param, value: os.path.abspath(os.path.expanduser(value)),
)
@click.option(
    "--output-dir",
    "-o",
    type=click.Path(exists=False),
    help="Output directory to unpack the zip file. Defaults to the directory containing the zip file.",
    default=None,
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    default=False,
    help="Force overwrite if the output directory already exists.",
    show_default=True,
)
def deploy(zip_file: str, output_dir: str | None, force: bool):
    """
    Deploys a strategy from a zip file created by the release command.

    This command:
    1. Unpacks the zip file to the specified output directory
    2. Creates a Poetry virtual environment in the .venv folder
    3. Installs dependencies from the poetry.lock file

    If no output directory is specified, the zip file is unpacked in the same directory
    as the zip file, in a folder with the same name as the zip file (without the .zip extension).
    """
    from .deploy import deploy_strategy

    deploy_strategy(zip_file, output_dir, force)


if __name__ == "__main__":
    main()



from qubx.backtester.ome import OmeReport
from qubx.core.basics import (
    CtrlChannel,
    Instrument,
    Order,
)
from qubx.core.interfaces import IBroker

from .account import SimulatedAccountProcessor


class SimulatedBroker(IBroker):
    channel: CtrlChannel

    _account: SimulatedAccountProcessor

    def __init__(
        self,
        channel: CtrlChannel,
        account: SimulatedAccountProcessor,
        exchange_id: str = "simulated",
    ) -> None:
        self.channel = channel
        self._account = account
        self._exchange_id = exchange_id

    @property
    def is_simulated_trading(self) -> bool:
        return True

    def send_order(
        self,
        instrument: Instrument,
        order_side: str,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **options,
    ) -> Order:
        ome = self._account.ome.get(instrument)
        if ome is None:
            raise ValueError(f"ExchangeService:send_order :: No OME configured for '{instrument.symbol}'!")

        # - try to place order in OME
        report = ome.place_order(
            order_side.upper(),  # type: ignore
            order_type.upper(),  # type: ignore
            amount,
            price,
            client_id,
            time_in_force,
            **options,
        )

        self._send_exec_report(instrument, report)
        return report.order

    def cancel_order(self, order_id: str) -> Order | None:
        instrument = self._account.order_to_instrument.get(order_id)
        if instrument is None:
            raise ValueError(f"ExchangeService:cancel_order :: can't find order with id = '{order_id}'!")

        ome = self._account.ome.get(instrument)
        if ome is None:
            raise ValueError(f"ExchangeService:send_order :: No OME configured for '{instrument}'!")

        # - cancel order in OME and remove from the map to free memory
        order_update = ome.cancel_order(order_id)
        self._send_exec_report(instrument, order_update)

        return order_update.order

    def cancel_orders(self, instrument: Instrument) -> None:
        raise NotImplementedError("Not implemented yet")

    def update_order(self, order_id: str, price: float | None = None, amount: float | None = None) -> Order:
        raise NotImplementedError("Not implemented yet")

    def _send_exec_report(self, instrument: Instrument, report: OmeReport):
        self.channel.send((instrument, "order", report.order, False))
        if report.exec is not None:
            self.channel.send((instrument, "deals", [report.exec], False))

    def exchange(self) -> str:
        return self._exchange_id.upper()



def recognize_time(time): ...
def time_to_str(t: int, units: str = "ns") -> str: ...
def time_delta_to_str(d: int) -> str: ...
def recognize_timeframe(timeframe): ...
def prec_ceil(a: float, precision: int) -> float: ...
def prec_floor(a: float, precision: int) -> float: ...



class BaseError(Exception):
    pass


class ExchangeError(BaseError):
    pass


class BadRequest(ExchangeError):
    pass


class InvalidOrder(ExchangeError):
    pass


class OrderNotFound(InvalidOrder):
    pass


class NotSupported(ExchangeError):
    pass


class QueueTimeout(BaseError):
    pass


class StrategyExceededMaxNumberOfRuntimeFailuresError(Exception):
    pass


class SimulationError(Exception):
    pass


class SimulationConfigError(Exception):
    pass
